# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import functools
import typing as tp
from abc import abstractmethod
from types import MappingProxyType
from typing import Any

import jax
import jax.numpy as jnp
import jax.stages

from flax.experimental.nnx.nnx import (
  filterlib,
  rnglib,
  spmd,
  tracers,
  variables,
)
from flax.experimental.nnx.nnx.module import GraphDef, Module, ModuleMeta
from flax.experimental.nnx.nnx.proxy_caller import (
  CallableProxy,
  DelayedAccessor,
)
from flax.experimental.nnx.nnx.state import State

A = tp.TypeVar('A')
C = tp.TypeVar('C')
B = tp.TypeVar('B')
F = tp.TypeVar('F', bound=tp.Callable[..., tp.Any])
G = tp.TypeVar('G', bound=tp.Callable[..., tp.Any])
M = tp.TypeVar('M', bound=Module)
N = tp.TypeVar('N', bound=Module)

AxisName = tp.Hashable
Leaf = tp.Any
Leaves = tp.List[Leaf]


def _check_args(args: tuple[tp.Any, ...]):
  """Check if Rngs is passed as a positional argument and raise an error."""
  for arg in args:
    if isinstance(arg, rnglib.Rngs):
      raise ValueError(
        "Rngs must be passed as a keyword argument named 'rngs', not a"
        ' positional argument'
      )


class LiftedModule(Module, tp.Generic[M]):
  @abstractmethod
  def _call(self, accessor: DelayedAccessor, *args, **kwargs) -> tp.Any:
    ...

  @property
  @abstractmethod
  def _submodule(self) -> M:
    ...

  def __call__(self, *args, **kwargs) -> tp.Any:
    return self.call(*args, **kwargs)  # type: ignore

  @property
  def call(self) -> tp.Any:
    module = self

    def check_and_call(accessor: DelayedAccessor, *args, **kwargs):
      _check_args(args)
      return self._call(accessor, *args, **kwargs)

    proxy = CallableProxy(check_and_call)

    while isinstance(module._submodule, LiftedModule):
      module = module._submodule
      proxy = proxy.call

    return proxy  # type: ignore


# -------------------------------
# jit
# -------------------------------

UNSPECIFIED = object()


@dataclasses.dataclass
class JITOptions:
  in_shardings: tp.Any
  out_shardings: tp.Any
  static_argnums: tp.Union[int, tp.Sequence[int], None]
  static_argnames: tp.Union[str, tp.Iterable[str], None]
  donate_argnums: tp.Union[int, tp.Sequence[int]]
  keep_unused: bool
  device: tp.Optional[jax.Device]
  backend: tp.Optional[str]
  inline: bool
  abstracted_axes: tp.Optional[tp.Any]

  def get_kwargs(self) -> dict[str, tp.Any]:
    kwargs = vars(self).copy()
    if kwargs['in_shardings'] is UNSPECIFIED:
      kwargs.pop('in_shardings')
    if kwargs['out_shardings'] is UNSPECIFIED:
      kwargs.pop('out_shardings')
    return kwargs


class JITMeta(ModuleMeta):
  def __call__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    in_shardings: tp.Any = UNSPECIFIED,
    out_shardings: tp.Any = UNSPECIFIED,
    static_argnums: tp.Union[int, tp.Sequence[int], None] = None,
    static_argnames: tp.Union[str, tp.Iterable[str], None] = None,
    donate_argnums: tp.Union[int, tp.Sequence[int]] = (),
    keep_unused: bool = False,
    device: tp.Optional[jax.Device] = None,
    backend: tp.Optional[str] = None,
    inline: bool = False,
    abstracted_axes: tp.Optional[tp.Any] = None,
  ) -> tp.Callable[..., 'JIT[M]']:
    super_call = super().__call__

    def _create_jit(*args, **kwargs) -> JIT[M]:
      _check_args(args)
      return super_call(
        module_constructor=module_constructor,
        in_shardings=in_shardings,
        out_shardings=out_shardings,
        static_argnums=static_argnums,
        static_argnames=static_argnames,
        donate_argnums=donate_argnums,
        keep_unused=keep_unused,
        device=device,
        backend=backend,
        inline=inline,
        abstracted_axes=abstracted_axes,
        # submodule args
        module_init_args=args,
        module_init_kwargs=kwargs,
      )

    return _create_jit


class JittedFn(tp.Protocol, tp.Generic[M]):
  def __call__(
    self, state_and_def: tuple[State | tuple[State, ...], GraphDef[M]]
  ) -> tuple[tuple[State | tuple[State, ...], GraphDef[M]], tp.Any]:
    ...


def get_jitted_fn(_module_type: type[M], f, options: JITOptions) -> JittedFn[M]:
  jit_kwargs = options.get_kwargs()

  @functools.partial(jax.jit, **jit_kwargs)
  def jitted_fn(
    state_and_def: tuple[State | tuple[State, ...], GraphDef[M]],
    *args,
    **kwargs,
  ):
    _check_args(args)
    states, graphdef = state_and_def

    if isinstance(states, State):
      states = (states,)

    nnx_trace = tracers.get_top_trace((args, kwargs))
    with tracers.nnx_trace(nnx_trace):
      if 'rngs' in kwargs:
        kwargs['rngs'] = rnglib.Rngs(kwargs['rngs'])
      module = graphdef.merge(*states)
      out = f(module, *args, **kwargs)

    updates = module.split()
    out = (updates, out)

    return out

  return jitted_fn


def jit_init(
  jitted_fn: JittedFn[M],
  module: M,
  args: tuple[tp.Any, ...],
  kwargs: dict[str, tp.Any],
) -> None:
  if not isinstance(module, Module):
    raise TypeError(f'Expected Module, got {type(module).__name__}')

  module = tp.cast(M, module)

  if 'rngs' in kwargs and isinstance(rngs := kwargs['rngs'], rnglib.Rngs):
    kwargs['rngs'] = rngs.fork()

  state_and_def = module.split()
  out = jitted_fn(state_and_def, *args, **kwargs)
  updates, _ = out
  module.update(updates)


def jit_apply(
  jitted_fn: JittedFn[M],
  module: M,
  args: tuple[tp.Any, ...],
  kwargs: dict[str, tp.Any],
) -> tp.Any:
  if not isinstance(module, Module):
    raise TypeError(f'Expected Module, got {type(module).__name__}')

  module = tp.cast(M, module)

  if 'rngs' in kwargs and isinstance(rngs := kwargs['rngs'], rnglib.Rngs):
    kwargs['rngs'] = rngs.fork()

  state_and_def = module.split()
  updates, out = jitted_fn(state_and_def, *args, **kwargs)
  module.update(updates)
  return out


class JIT(LiftedModule[M], metaclass=JITMeta):
  def __init__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    in_shardings: tp.Any = UNSPECIFIED,
    out_shardings: tp.Any = UNSPECIFIED,
    static_argnums: tp.Union[int, tp.Sequence[int], None] = None,
    static_argnames: tp.Union[str, tp.Iterable[str], None] = None,
    donate_argnums: tp.Union[int, tp.Sequence[int]] = (),
    keep_unused: bool = False,
    device: tp.Optional[jax.Device] = None,
    backend: tp.Optional[str] = None,
    inline: bool = False,
    abstracted_axes: tp.Optional[tp.Any] = None,
    # submodule args
    module_init_args: tuple[tp.Any, ...],
    module_init_kwargs: dict[str, tp.Any],
  ):
    self.options = JITOptions(
      in_shardings=in_shardings,
      out_shardings=out_shardings,
      static_argnums=static_argnums,
      static_argnames=static_argnames,
      donate_argnums=donate_argnums,
      keep_unused=keep_unused,
      device=device,
      backend=backend,
      inline=inline,
      abstracted_axes=abstracted_axes,
    )
    self.accessor: tp.Optional[DelayedAccessor] = None

    def jit_call_module(module, *args, **kwargs):
      assert self.accessor is not None
      f = self.accessor(module)
      return f(*args, **kwargs)

    self.jitted_fn: JittedFn[M] = get_jitted_fn(
      M, jit_call_module, self.options
    )
    self.module_constructor = module_constructor
    self.jit_module = self.module_constructor(
      *module_init_args, **module_init_kwargs
    )

  @property
  def _submodule(self) -> M:
    return self.jit_module

  def _call(self, accessor: DelayedAccessor, *args, **kwargs) -> Any:
    self.accessor = accessor
    try:
      out = jit_apply(self.jitted_fn, self.jit_module, args, kwargs)
    finally:
      self.accessor = None
    return out


def jit(
  f: F,
  *,
  in_shardings: tp.Any = UNSPECIFIED,
  out_shardings: tp.Any = UNSPECIFIED,
  static_argnums: tp.Union[int, tp.Sequence[int], None] = None,
  static_argnames: tp.Union[str, tp.Iterable[str], None] = None,
  donate_argnums: tp.Union[int, tp.Sequence[int]] = (),
  keep_unused: bool = False,
  device: tp.Optional[jax.Device] = None,
  backend: tp.Optional[str] = None,
  inline: bool = False,
  abstracted_axes: tp.Optional[tp.Any] = None,
  is_init: tp.Optional[bool] = None,
) -> F:
  if is_init is None:
    is_init = f.__name__ == '__init__'

  if static_argnames is None:
    static_argnames = []
  elif isinstance(static_argnames, str):
    static_argnames = [static_argnames]
  else:
    static_argnames = list(static_argnames)

  options = JITOptions(
    in_shardings=in_shardings,
    out_shardings=out_shardings,
    static_argnums=static_argnums,
    static_argnames=static_argnames,
    donate_argnums=donate_argnums,
    keep_unused=keep_unused,
    device=device,
    backend=backend,
    inline=inline,
    abstracted_axes=abstracted_axes,
  )
  jitted_fn = get_jitted_fn(Module, f, options)

  if is_init:

    @functools.wraps(f)
    def jit_init_wrapper(module: Module, *args, **kwargs):
      _check_args(args)
      jit_init(jitted_fn, module, args, kwargs)

    wrapper = jit_init_wrapper
    wrapper.inner = jitted_fn
  else:

    @functools.wraps(f)
    def jit_apply_wrapper(module: Module, *args, **kwargs):
      _check_args(args)
      return jit_apply(jitted_fn, module, args, kwargs)

    wrapper = jit_apply_wrapper
    wrapper.inner = jitted_fn

  return wrapper  # type: ignore


# -------------------------------
# grad
# -------------------------------


@dataclasses.dataclass
class GradOptions:
  wrt: filterlib.Filter
  has_aux: bool
  holomorphic: bool
  allow_int: bool
  reduce_axes: tp.Sequence[AxisName]
  return_value: bool


class GradMeta(ModuleMeta):
  def __call__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    wrt: filterlib.Filter = variables.Param,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
    return_value: bool = False,
  ) -> tp.Callable[..., 'Grad[M]']:
    super_call = super().__call__

    def _create_grad(*args, **kwargs) -> Grad[M]:
      _check_args(args)
      return super_call(
        module_constructor=module_constructor,
        wrt=wrt,
        has_aux=has_aux,
        holomorphic=holomorphic,
        allow_int=allow_int,
        reduce_axes=reduce_axes,
        return_value=return_value,
        # submodule args
        module_init_args=args,
        module_init_kwargs=kwargs,
      )

    return _create_grad


class Grad(LiftedModule[M], metaclass=GradMeta):
  def __init__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    wrt: filterlib.Filter = variables.Param,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
    return_value: bool = False,
    # submodule args
    module_init_args: tuple[tp.Any, ...],
    module_init_kwargs: dict[str, tp.Any],
  ):
    self.options = GradOptions(
      wrt=wrt,
      has_aux=has_aux,
      holomorphic=holomorphic,
      allow_int=allow_int,
      reduce_axes=reduce_axes,
      return_value=return_value,
    )
    self.module_constructor = module_constructor
    self.grad_module = self.module_constructor(
      *module_init_args, **module_init_kwargs
    )

  @property
  def _submodule(self) -> M:
    return self.grad_module

  def _call(self, accessor: DelayedAccessor, *args, **kwargs) -> Any:
    def grad_call_apply(module, *args, **kwargs):
      return accessor(module)(*args, **kwargs)

    return grad_apply(
      self.options, grad_call_apply, self.grad_module, *args, **kwargs
    )


def grad_apply(options: GradOptions, f, module: Module, *args, **kwargs):
  if not isinstance(module, Module):
    raise TypeError(f'Expected a Module, got {type(module).__name__}')

  predicate = filterlib.to_predicate(options.wrt)

  diff, nondiff, graphdef = module.split(predicate, ...)
  transform = jax.value_and_grad if options.return_value else jax.grad

  @functools.partial(
    transform,
    argnums=0,  # we'll handle this ourselves
    has_aux=True,
    holomorphic=options.holomorphic,
    allow_int=options.allow_int,
    reduce_axes=options.reduce_axes,
  )
  def grad_fn(diff: State):
    nonlocal graphdef

    with tracers.nnx_trace(tracers.get_top_trace(diff)):
      module = graphdef.merge(diff, nondiff)
      out = f(module, *args, **kwargs)

    updates, graphdef = module.split()
    if options.has_aux:
      loss, aux = out
      out = (loss, (updates, aux))
    else:
      out = (out, updates)

    return out

  out = grad_fn(diff)

  updates: State
  if options.return_value:
    if options.has_aux:
      (loss, (updates, aux)), grads = out
      out = (loss, aux), grads
    else:
      (loss, updates), grads = out
      out = loss, grads
  else:
    if options.has_aux:
      grads, (updates, aux) = out
      out = grads, aux
    else:
      out, updates = out

  module.update(updates, graphdef)
  return out


@tp.overload
def grad(
  f: tp.Callable[..., tp.Any],
  wrt: filterlib.Filter = variables.Param,
  *,
  holomorphic: bool = False,
  allow_int: bool = False,
  reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., State]:
  ...


@tp.overload
def grad(
  f: tp.Callable[..., tp.Any],
  wrt: filterlib.Filter = variables.Param,
  *,
  has_aux: tp.Literal[True],
  holomorphic: bool = False,
  allow_int: bool = False,
  reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tuple[State, tp.Any]]:
  ...


def grad(
  f: tp.Callable[..., tp.Any],
  wrt: filterlib.Filter = variables.Param,
  *,
  has_aux: bool = False,
  holomorphic: bool = False,
  allow_int: bool = False,
  reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tp.Union[tuple[State, tp.Any], State]]:
  if f.__name__ == '__init__':
    raise ValueError('Cannot use `grad` with `__init__`')

  options = GradOptions(
    wrt=wrt,
    has_aux=has_aux,
    holomorphic=holomorphic,
    allow_int=allow_int,
    reduce_axes=reduce_axes,
    return_value=False,
  )

  @functools.wraps(f)
  def grad_wrapper(module: Module, *args, **kwargs):
    _check_args(args)
    return grad_apply(options, f, module, *args, **kwargs)

  return grad_wrapper  # type: ignore


@tp.overload
def value_and_grad(
  f: tp.Callable[..., tp.Any],
  wrt: filterlib.Filter = variables.Param,
  *,
  holomorphic: bool = False,
  allow_int: bool = False,
  reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tuple[jax.Array, State]]:
  ...


@tp.overload
def value_and_grad(
  f: tp.Callable[..., tp.Any],
  wrt: filterlib.Filter = variables.Param,
  *,
  has_aux: tp.Literal[True],
  holomorphic: bool = False,
  allow_int: bool = False,
  reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tuple[tuple[jax.Array, tp.Any], State]]:
  ...


def value_and_grad(
  f: tp.Callable[..., tp.Any],
  wrt: filterlib.Filter = variables.Param,
  *,
  has_aux: bool = False,
  holomorphic: bool = False,
  allow_int: bool = False,
  reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[
  ...,
  tp.Union[tuple[tuple[jax.Array, tp.Any], State], tuple[jax.Array, State]],
]:
  if f.__name__ == '__init__':
    raise ValueError('Cannot use `value_and_grad` with `__init__`')

  options = GradOptions(
    wrt=wrt,
    has_aux=has_aux,
    holomorphic=holomorphic,
    allow_int=allow_int,
    reduce_axes=reduce_axes,
    return_value=True,
  )

  @functools.wraps(f)
  def value_and_grad_wrapper(module: Module, *args, **kwargs):
    _check_args(args)
    return grad_apply(options, f, module, *args, **kwargs)

  return value_and_grad_wrapper  # type: ignore


# -------------------------------
# scan
# -------------------------------


@dataclasses.dataclass
class ScanOptions:
  variable_axes: tp.Mapping[filterlib.Filter, int]
  broadcast_rngs: filterlib.Filter
  in_args_axes: tp.Any
  in_kwargs_axes: tp.Any
  out_axes: tp.Any
  length: tp.Optional[int]
  reverse: bool
  unroll: int
  scan_metadata: tp.Mapping[str, tp.Any]
  scan_output: bool


class ScanMeta(ModuleMeta):
  def __call__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    variable_axes: tp.Mapping[filterlib.Filter, int] = MappingProxyType({}),
    broadcast_rngs: filterlib.Filter = None,
    in_args_axes: tp.Any = 0,
    in_kwargs_axes: tp.Any = 0,
    out_axes: tp.Any = 0,
    length: tp.Optional[int] = None,
    reverse: bool = False,
    unroll: int = 1,
    scan_metadata: tp.Mapping[str, tp.Any] = MappingProxyType({}),
    scan_output: bool = True,
  ) -> tp.Callable[..., 'Scan[M]']:
    super_call = super().__call__

    def _create_scan(*args, **kwargs) -> Scan[M]:
      _check_args(args)
      return super_call(
        module_constructor=module_constructor,
        module_init_args=args,
        module_init_kwargs=kwargs,
        variable_axes=variable_axes,
        broadcast_rngs=broadcast_rngs,
        in_args_axes=in_args_axes,
        in_kwargs_axes=in_kwargs_axes,
        out_axes=out_axes,
        length=length,
        reverse=reverse,
        unroll=unroll,
        scan_metadata=scan_metadata,
        scan_output=scan_output,
      )

    return _create_scan


class Scan(LiftedModule[M], metaclass=ScanMeta):
  def __init__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    variable_axes: tp.Mapping[filterlib.Filter, int] = MappingProxyType({}),
    broadcast_rngs: filterlib.Filter = None,
    in_args_axes: tp.Any = 0,
    in_kwargs_axes: tp.Any = 0,
    out_axes: tp.Any = 0,
    length: tp.Optional[int] = None,
    reverse: bool = False,
    unroll: int = 1,
    scan_metadata: tp.Mapping[str, tp.Any] = MappingProxyType({}),
    scan_output: bool = True,
    # submodule args
    module_init_args: tuple[tp.Any, ...],
    module_init_kwargs: dict[str, tp.Any],
  ):
    self.module_constructor = module_constructor
    self.options = ScanOptions(
      variable_axes=variable_axes,
      broadcast_rngs=broadcast_rngs,
      in_args_axes=in_args_axes,
      in_kwargs_axes=in_kwargs_axes,
      out_axes=out_axes,
      length=length,
      reverse=reverse,
      unroll=unroll,
      scan_metadata=scan_metadata,
      scan_output=scan_output,
    )
    self.scan_module = scan_init(
      self.options, module_constructor, module_init_args, module_init_kwargs
    )

  @property
  def _submodule(self) -> M:
    return self.scan_module

  def _call(
    self, accessor: DelayedAccessor, *args, **kwargs
  ) -> tuple[tp.Any, tp.Any]:
    if len(args) < 1:
      raise TypeError(
        f'Expected at least 1 positional arguments, got {len(args)}'
      )
    _check_args(args)
    carry_arg, args = args[0], args[1:]

    def scan_call_apply(module, *args, **kwargs):
      return accessor(module)(*args, **kwargs)

    return scan_apply(
      self.options,
      scan_call_apply,
      self.scan_module,
      carry_arg,
      args,
      kwargs,
    )


class ScanCall(tp.Protocol, tp.Generic[C, B]):
  def __call__(
    self,
    module: Module,
    carry_arg: C,
    *args: tp.Any,
    **kwargs: tp.Any,
  ) -> tuple[C, B] | C:
    ...


def scan_init(
  options: ScanOptions,
  module_constructor: tp.Callable[..., M],
  module_init_args: tuple[tp.Any, ...],
  module_init_kwargs: dict[str, tp.Any],
) -> M:
  if options.variable_axes and options.length is None:
    raise ValueError('Cannot use variable_axes without specifying a length')

  _check_args(module_init_args)

  rngs = module_init_kwargs.pop('rngs', None)

  if rngs is not None and not isinstance(rngs, rnglib.Rngs):
    raise TypeError(f'Expected a Rngs, got {type(rngs).__name__}')

  split_keys = []

  if rngs is not None:
    if not isinstance(rngs, rnglib.Rngs):
      raise TypeError(f'Expected a Rngs, got {type(rngs).__name__}')

    split_keys, broadcast_keys = rngs.fork(
      {filterlib.Not(options.broadcast_rngs): options.length}
    )

    if split_keys and options.length is None:
      raise ValueError('Cannot split RNGs without specifying a length')

  else:
    split_keys = None
    broadcast_keys = None

  graphdef: tp.Optional[GraphDef[M]] = None

  def _init_state(split_keys, broadcast_keys):
    nonlocal graphdef

    if split_keys is not None:
      assert broadcast_keys is not None
      module_init_kwargs['rngs'] = rnglib.Rngs(**split_keys, **broadcast_keys)

    module = module_constructor(*module_init_args, **module_init_kwargs)

    # lift module
    filters = (*options.variable_axes.keys(), ...)

    *states, graphdef = module.split(*filters)

    return tuple(states)

  if split_keys is not None or options.variable_axes:
    init_out_axes = (*options.variable_axes.values(), None)
    _init_state = jax.vmap(
      _init_state,
      in_axes=(0, None),
      out_axes=init_out_axes,
      axis_size=options.length,
    )

  *axes_states, carry_state = _init_state(split_keys, broadcast_keys)
  graphdef = tp.cast(GraphDef[M], graphdef)

  # add additional axis name to Variable.sharding
  if spmd.PARTITION_NAME in options.scan_metadata:
    axes_states = [
      spmd.add_axis(state, index, options.scan_metadata)
      for state, index in zip(axes_states, options.variable_axes.values())
    ]

  module = graphdef.merge(*axes_states, carry_state)

  return module


def scan_apply(
  options: ScanOptions,
  f: ScanCall[C, B],
  module: Module,
  carry_arg: C,
  args: tuple[tp.Any, ...],
  kwargs: dict[str, tp.Any],
) -> tuple[C, B] | C:
  rngs = kwargs.pop('rngs', None)

  # split module state
  filters = (*options.variable_axes.keys(), ...)
  *scan_states, carry_state, graphdef = module.split(*filters)

  # transpose axes state
  scan_states = tuple(
    jax.tree_map(lambda x: jnp.moveaxis(x, axis, 0), axes_state)
    for axes_state, axis in zip(scan_states, options.variable_axes.values())
  )
  # transpose axes arg
  scan_args = jax.tree_map(
    lambda axis, node: jax.tree_map(lambda x: jnp.moveaxis(x, axis, 0), node)
    if axis is not None
    else None,
    options.in_args_axes,
    args,
    is_leaf=lambda x: x is None,
  )
  broadcast_args = jax.tree_map(
    lambda axis, node: None if axis is not None else node,
    options.in_args_axes,
    args,
    is_leaf=lambda x: x is None,
  )
  scan_kwargs = jax.tree_map(
    lambda axis, node: jax.tree_map(lambda x: jnp.moveaxis(x, axis, 0), node)
    if axis is not None
    else None,
    options.in_kwargs_axes,
    kwargs,
    is_leaf=lambda x: x is None,
  )
  broadcast_kwargs = jax.tree_map(
    lambda axis, node: None if axis is not None else node,
    options.in_kwargs_axes,
    kwargs,
    is_leaf=lambda x: x is None,
  )

  # infer length
  lengths: tp.Set[int] = set(
    x.shape[0]
    for x in jax.tree_util.tree_leaves((scan_states, scan_args, scan_kwargs))
  )

  if len(lengths) > 1:
    raise ValueError(
      'Inconsistent lengths between variable_axes states and '
      f'arguments: {lengths}'
    )
  elif len(lengths) == 0:
    if options.length is None:
      raise ValueError(
        'Cannot infer length from variable_axes states or axes_arg, '
        'please specify `length`'
      )
    length = options.length
  else:
    length = lengths.pop()
    if options.length is not None and options.length != length:
      raise ValueError(
        f'Specified length {options.length} is not the same as the inferred '
        f'length {length}'
      )

  # split rng state
  if rngs is not None:
    if not isinstance(rngs, rnglib.Rngs):
      raise TypeError(f'Expected a Rngs, got {type(rngs).__name__}')
    split_keys, broadcast_keys = rngs.fork(
      {filterlib.Not(options.broadcast_rngs): length}
    )
  else:
    split_keys = None
    broadcast_keys = None

  moduledef_out: tp.Optional[GraphDef[Module]] = None

  def scan_fn(
    carry: tuple[State, tp.Any],
    scan: tuple[
      dict[str, rnglib.RngStream] | None,
      tuple[State, ...],
      tuple[tp.Any, ...],
      dict[str, tp.Any],
    ],
  ):
    nonlocal moduledef_out
    carry_state, carry_arg = carry
    split_keys, scan_states, scan_args, scan_kwargs = scan

    # merge args and kwargs
    args = jax.tree_map(
      lambda axis, scan, broadcast: scan if axis is not None else broadcast,
      options.in_args_axes,
      scan_args,
      broadcast_args,
      is_leaf=lambda x: x is None,
    )
    kwargs = jax.tree_map(
      lambda axis, scan, broadcast: scan if axis is not None else broadcast,
      options.in_kwargs_axes,
      scan_kwargs,
      broadcast_kwargs,
      is_leaf=lambda x: x is None,
    )

    # merge rng state
    if split_keys is not None:
      assert broadcast_keys is not None
      kwargs['rngs'] = rnglib.Rngs(**split_keys, **broadcast_keys)

    # remove metadata axis name from Variable.sharding
    if spmd.PARTITION_NAME in options.scan_metadata:
      scan_states = [
        spmd.remove_axis(state, index, options.scan_metadata)
        for state, index in zip(scan_states, options.variable_axes.values())
      ]

    # merge module state
    module = graphdef.merge(*scan_states, carry_state)

    output = f(module, carry_arg, *args, **kwargs)

    if options.scan_output:
      if not isinstance(output, tuple) or len(output) != 2:
        raise ValueError(
          'Expected a tuple of length 2 as the output of the scan function, '
          f'got {output}'
        )
      output = tp.cast(tuple[C, B], output)
      carry_out, scan_out = output
    else:
      output = tp.cast(C, output)
      carry_out = output
      scan_out = None

    # split module state
    *scan_states_out, carry_state_out, moduledef_out = module.split(*filters)
    carry_state_new = carry_state_out - carry_state

    # remove new carry state
    carry_state_out = carry_state_out - carry_state_new

    # add metadata axis name to Variable.sharding
    if spmd.PARTITION_NAME in options.scan_metadata:
      scan_states_out = [
        spmd.add_axis(state, index, options.scan_metadata)
        for state, index in zip(scan_states_out, options.variable_axes.values())
      ]

    full_carry_out = (carry_state_out, carry_out)
    full_scan_out = (scan_states_out, carry_state_new, scan_out)

    return full_carry_out, full_scan_out

  carry = (carry_state, carry_arg)
  scan = (split_keys, scan_states, scan_args, scan_kwargs)

  full_carry_out, full_scan_out = jax.lax.scan(
    scan_fn,
    carry,
    scan,
    length=length,
    reverse=options.reverse,
    unroll=options.unroll,
  )
  carry_state, carry_out = full_carry_out
  scan_states, carry_state_new, scan_out = full_scan_out
  assert moduledef_out is not None

  # transpose axes state
  scan_states = tuple(
    jax.tree_map(lambda x: jnp.moveaxis(x, 0, axis), axes_state)
    for axes_state, axis in zip(scan_states, options.variable_axes.values())
  )
  # transpose axes arg
  scan_out = jax.tree_map(
    lambda axis, node: jax.tree_map(lambda x: jnp.moveaxis(x, 0, axis), node),
    options.out_axes,
    scan_out,
  )
  # slice new carry state
  carry_state_new = jax.tree_map(lambda x: x[0], carry_state_new)

  module.update(((*scan_states, carry_state, carry_state_new), moduledef_out))

  if options.scan_output:
    return carry_out, scan_out
  else:
    return carry_out


def scan(
  f: F,
  *,
  variable_axes: tp.Mapping[filterlib.Filter, int] = MappingProxyType({}),
  broadcast_rngs: filterlib.Filter = None,
  in_args_axes: tp.Any = 0,
  in_kwargs_axes: tp.Any = 0,
  out_axes: tp.Any = 0,
  length: tp.Optional[int] = None,
  reverse: bool = False,
  unroll: int = 1,
  is_init: tp.Optional[bool] = None,
  scan_metadata: tp.Mapping[str, tp.Any] = {},
  scan_output: bool = True,
) -> F:
  if is_init is None:
    is_init = f.__name__ == '__init__'

  options = ScanOptions(
    variable_axes=variable_axes,
    broadcast_rngs=broadcast_rngs,
    in_args_axes=in_args_axes,
    in_kwargs_axes=in_kwargs_axes,
    out_axes=out_axes,
    length=length,
    reverse=reverse,
    unroll=unroll,
    scan_metadata=scan_metadata,
    scan_output=scan_output,
  )

  if is_init:

    @functools.wraps(f)
    def scan_init_wrapper(module: Module, *args, **kwargs):
      def module_constructor(*args, **kwargs):
        _check_args(args)
        f(module, *args, **kwargs)
        return module

      lifted_module = scan_init(options, module_constructor, args, kwargs)
      module.update(lifted_module)

    wrapper = scan_init_wrapper

  else:

    @functools.wraps(f)
    def scan_apply_wrapper(
      module: Module,
      *args,
      **kwargs,
    ) -> tuple[C, tp.Any]:
      if len(args) < 2:
        raise TypeError(
          f'Expected at least 2 positional arguments, got {len(args)}'
        )
      _check_args(args)

      carry_arg, args = args[0], args[1:]
      return scan_apply(options, f, module, carry_arg, args, kwargs)

    wrapper = scan_apply_wrapper

  return wrapper  # type: ignore


# -------------------------------
# remat
# -------------------------------


class RematMeta(ModuleMeta):
  def __call__(
    self,
    module_constructor: tp.Callable[..., M],
    # variables: lift.CollectionFilter = True,
    # rngs: lift.PRNGSequenceFilter = True,
    prevent_cse: bool = True,
    static_argnums: tp.Union[int, tuple[int, ...]] = (),
    policy: tp.Optional[tp.Callable[..., bool]] = None,
  ) -> tp.Callable[..., 'Remat[M]']:
    super_call = super().__call__

    def create_remat(*args, **kwargs) -> Remat[M]:
      _check_args(args)
      return super_call(
        module_constructor=module_constructor,
        module_init_args=args,
        module_init_kwargs=kwargs,
        prevent_cse=prevent_cse,
        static_argnums=static_argnums,
        policy=policy,
      )

    return create_remat


@dataclasses.dataclass
class RematOptions:
  prevent_cse: bool
  static_argnums: tp.Union[int, tuple[int, ...]]
  policy: tp.Optional[tp.Callable[..., bool]]

  def __post_init__(self):
    if isinstance(self.static_argnums, int):
      self.static_argnums = (self.static_argnums,)

    # add 2 as an offset to account for state and keys
    self.static_argnums = tuple(
      x + 2 if x >= 0 else x for x in self.static_argnums
    )


class Remat(LiftedModule[M], metaclass=RematMeta):
  def __init__(
    self,
    *,
    module_constructor: tp.Callable[..., M],
    prevent_cse: bool = True,
    static_argnums: tp.Union[int, tuple[int, ...]] = (),
    policy: tp.Optional[tp.Callable[..., bool]] = None,
    # submodule args
    module_init_args: tuple[tp.Any, ...],
    module_init_kwargs: dict[str, tp.Any],
  ):
    self.options = RematOptions(
      prevent_cse=prevent_cse,
      static_argnums=static_argnums,
      policy=policy,
    )
    self.module_constructor = module_constructor
    self.remat_module = self.module_constructor(
      *module_init_args, **module_init_kwargs
    )

  @property
  def _submodule(self) -> M:
    return self.remat_module

  def _call(
    self,
    accessor: DelayedAccessor,
    *args,
    rngs: tp.Optional[rnglib.Rngs] = None,
  ) -> tp.Any:
    def remat_call_apply(module, *args, **kwargs):
      return accessor(module)(*args, **kwargs)

    return remat_apply(
      self.options,
      remat_call_apply,
      self.remat_module,
      args,
      rngs,
    )


class RematCall(tp.Protocol):
  def __call__(self, *args, rngs: tp.Optional[rnglib.Rngs]) -> tp.Any:
    ...


def remat_apply(
  options: RematOptions,
  f: RematCall,
  module: Module,
  args: tuple[tp.Any, ...],
  rngs: tp.Optional[rnglib.Rngs],
):
  _check_args(args)

  state, graphdef = module.split()
  keys = rngs.fork() if rngs is not None else None

  def _remat_fn(
    state: State,
    keys: tp.Optional[dict[str, jax.Array]],
    *args,
  ) -> tuple[tuple[State, GraphDef[Module]], tp.Any]:
    kwargs = {}
    if keys is not None:
      kwargs['rngs'] = rnglib.Rngs(keys)

    module = graphdef.merge(state)
    out = f(module, *args, **kwargs)

    state_and_def = module.split()

    return state_and_def, out

  state_and_def: tuple[State, GraphDef[Module]]
  state_and_def, out = jax.checkpoint(
    _remat_fn,
    prevent_cse=options.prevent_cse,
    static_argnums=options.static_argnums,
    policy=options.policy,
  )(state, keys, *args)

  module.update(state_and_def)

  return out


def remat(
  f: F,
  *,
  # variables: lift.CollectionFilter,
  # rngs: lift.PRNGSequenceFilter,
  prevent_cse: bool = True,
  static_argnums: tp.Union[int, tuple[int, ...]] = (),
  policy: tp.Optional[tp.Callable[..., bool]] = None,
  is_init: tp.Optional[bool] = None,
) -> F:
  if is_init is None:
    is_init = f.__name__ == '__init__'

  options = RematOptions(
    # variables=variables,
    # rngs=rngs,
    prevent_cse=prevent_cse,
    static_argnums=static_argnums,
    policy=policy,
  )

  if is_init:
    return f
  else:

    @functools.wraps(f)
    def remat_wrapper(
      module: Module, *args, rngs: tp.Optional[rnglib.Rngs] = None
    ):
      return remat_apply(options, f, module, args, rngs)

    return remat_wrapper  # type: ignore


# -------------------------------
# vmap
# -------------------------------


@dataclasses.dataclass
class VmapOptions:
  variable_axes: tp.Mapping[filterlib.Filter, int]
  broadcast_rngs: filterlib.Filter
  in_args_axes: tp.Any
  in_kwargs_axes: tp.Any
  out_axes: tp.Any
  axis_size: int | None
  axis_name: str | None
  spmd_axis_name: str | None
  vmap_metadata: tp.Mapping[str, tp.Any]


class VmapMeta(ModuleMeta):
  def __call__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    variable_axes: tp.Mapping[filterlib.Filter, int] = MappingProxyType({}),
    broadcast_rngs: filterlib.Filter = None,
    in_args_axes: tp.Any = 0,
    in_kwargs_axes: tp.Any = 0,
    out_axes: tp.Any = 0,
    axis_size: int | None = None,
    axis_name: str | None = None,
    spmd_axis_name: str | None = None,
    vmap_metadata: tp.Mapping[str, tp.Any] = MappingProxyType({}),
  ) -> tp.Callable[..., 'Vmap[M]']:
    super_call = super().__call__

    def _create_scan(*args, **kwargs) -> Scan[M]:
      _check_args(args)
      return super_call(
        module_constructor=module_constructor,
        module_init_args=args,
        module_init_kwargs=kwargs,
        variable_axes=variable_axes,
        broadcast_rngs=broadcast_rngs,
        in_args_axes=in_args_axes,
        in_kwargs_axes=in_kwargs_axes,
        out_axes=out_axes,
        axis_size=axis_size,
        axis_name=axis_name,
        spmd_axis_name=spmd_axis_name,
        vmap_metadata=vmap_metadata,
      )

    return _create_scan


class Vmap(LiftedModule[M], metaclass=VmapMeta):
  def __init__(
    self,
    module_constructor: tp.Callable[..., M],
    *,
    variable_axes: tp.Mapping[filterlib.Filter, int] = MappingProxyType({}),
    broadcast_rngs: filterlib.Filter = None,
    in_args_axes: tp.Any = 0,
    in_kwargs_axes: tp.Any = 0,
    out_axes: tp.Any = 0,
    axis_size: int | None = None,
    axis_name: str | None = None,
    spmd_axis_name: str | None = None,
    vmap_metadata: tp.Mapping[str, tp.Any] = MappingProxyType({}),
    # submodule args
    module_init_args: tuple[tp.Any, ...],
    module_init_kwargs: dict[str, tp.Any],
  ):
    self.module_constructor = module_constructor
    self.options = VmapOptions(
      variable_axes=variable_axes,
      broadcast_rngs=broadcast_rngs,
      in_args_axes=in_args_axes,
      in_kwargs_axes=in_kwargs_axes,
      out_axes=out_axes,
      axis_size=axis_size,
      axis_name=axis_name,
      spmd_axis_name=spmd_axis_name,
      vmap_metadata=vmap_metadata,
    )
    self.vmap_module = vmap_init(
      self.options, module_constructor, module_init_args, module_init_kwargs
    )

  @property
  def _submodule(self) -> M:
    return self.vmap_module

  def _call(
    self, accessor: DelayedAccessor, *args, **kwargs
  ) -> tuple[tp.Any, tp.Any]:
    _check_args(args)

    def vmap_call_apply(module, *args, **kwargs):
      return accessor(module)(*args, **kwargs)

    return vmap_apply(
      self.options,
      vmap_call_apply,
      self.vmap_module,
      args,
      kwargs,
    )


class VmapCall(tp.Protocol):
  def __call__(
    self,
    module: Module,
    *args: tp.Any,
    **kwargs: tp.Any,
  ) -> tp.Any:
    ...


def vmap_init(
  options: VmapOptions,
  module_constructor: tp.Callable[..., M],
  module_init_args: tuple[tp.Any, ...],
  module_init_kwargs: dict[str, tp.Any],
) -> M:
  if options.variable_axes and options.axis_size is None:
    raise ValueError('Cannot use variable_axes without specifying a length')

  _check_args(module_init_args)

  rngs = module_init_kwargs.pop('rngs', None)

  if rngs is not None and not isinstance(rngs, rnglib.Rngs):
    raise TypeError(f'Expected a Rngs, got {type(rngs).__name__}')

  if rngs is not None:
    if not isinstance(rngs, rnglib.Rngs):
      raise TypeError(f'Expected a Rngs, got {type(rngs).__name__}')
    split_keys, broadcast_keys = rngs.fork(
      {filterlib.Not(options.broadcast_rngs): options.axis_size}
    )
    if split_keys and options.axis_size is None:
      raise ValueError('Cannot split RNGs without specifying a length')
  else:
    split_keys = None
    broadcast_keys = None

  graphdef: tp.Optional[GraphDef[M]] = None

  def _init_state(split_keys, broadcast_keys):
    nonlocal graphdef

    if split_keys is not None:
      assert broadcast_keys is not None
      module_init_kwargs['rngs'] = rnglib.Rngs(**split_keys, **broadcast_keys)

    module = module_constructor(*module_init_args, **module_init_kwargs)

    # lift module
    filters = (*options.variable_axes.keys(), ...)

    *states, graphdef = module.split(*filters)

    return tuple(states)

  if split_keys is not None or options.variable_axes:
    init_out_axes = (*options.variable_axes.values(), None)
    _init_state = jax.vmap(
      _init_state,
      in_axes=(0, None),
      out_axes=init_out_axes,
      axis_size=options.axis_size,
    )

  *axes_states, carry_state = _init_state(split_keys, broadcast_keys)
  graphdef = tp.cast(GraphDef[M], graphdef)

  # add additional axis name to Variable.sharding
  if spmd.PARTITION_NAME in options.vmap_metadata:
    axes_states = [
      spmd.add_axis(state, index, options.vmap_metadata)
      for state, index in zip(axes_states, options.variable_axes.values())
    ]

  module = graphdef.merge(*axes_states, carry_state)
  return module


def vmap_apply(
  options: VmapOptions,
  f: VmapCall,
  module: Module,
  args: tuple[tp.Any, ...],
  kwargs: dict[str, tp.Any],
) -> tp.Any:
  rngs = kwargs.pop('rngs', None)

  # split module state
  filters = (*options.variable_axes.keys(), ...)
  *vectorized_states, broadcast_state, graphdef = module.split(*filters)

  # infer length
  axis_sizes: tp.Set[int] = set()
  args_sizes = jax.tree_map(
    lambda axis, node: jax.tree_map(lambda x: x.shape[axis], node)
    if axis is not None
    else None,
    options.in_args_axes,
    args,
    is_leaf=lambda x: x is None,
  )
  kwargs_sizes = jax.tree_map(
    lambda axis, node: jax.tree_map(lambda x: x.shape[axis], node)
    if axis is not None
    else None,
    options.in_kwargs_axes,
    kwargs,
    is_leaf=lambda x: x is None,
  )
  axis_sizes.update(jax.tree_util.tree_leaves(args_sizes))
  axis_sizes.update(jax.tree_util.tree_leaves(kwargs_sizes))

  if len(axis_sizes) > 1:
    raise ValueError(
      'Inconsistent lengths between variable_axes states and '
      f'arguments: {axis_sizes}'
    )
  elif len(axis_sizes) == 0:
    if options.axis_size is None:
      raise ValueError(
        'Cannot infer length from variable_axes states or axes_arg, '
        'please specify `length`'
      )
    axis_size = options.axis_size
  else:
    axis_size = axis_sizes.pop()
    if options.axis_size is not None and options.axis_size != axis_size:
      raise ValueError(
        f'Specified axis_size {options.axis_size} is not the same as the'
        f' inferred length {axis_size}'
      )

  # split rng state
  if rngs is not None:
    if not isinstance(rngs, rnglib.Rngs):
      raise TypeError(f'Expected a Rngs, got {type(rngs).__name__}')

    split_keys, broadcast_keys = rngs.fork(
      {filterlib.Not(options.broadcast_rngs): axis_size}
    )
  else:
    split_keys = None
    broadcast_keys = None

  moduledef_out: tp.Optional[GraphDef[Module]] = None

  keys_axes = 0
  states_axes = list(options.variable_axes.values())
  args_axes = options.in_args_axes
  kwargs_axes = options.in_kwargs_axes
  out_axes = options.out_axes

  @functools.partial(
    jax.vmap,
    in_axes=(keys_axes, states_axes, args_axes, kwargs_axes),
    out_axes=(None, states_axes, out_axes),
    axis_name=options.axis_name,
    axis_size=axis_size,
    spmd_axis_name=options.spmd_axis_name,
  )
  def vmap_fn(
    split_keys: dict[str, rnglib.RngStream] | None,
    vectorized_states: list[State],
    args: tuple[tp.Any, ...],
    kwargs: dict[str, tp.Any],
  ):
    nonlocal moduledef_out

    # merge rng state
    if split_keys is not None:
      assert broadcast_keys is not None
      kwargs['rngs'] = rnglib.Rngs(**split_keys, **broadcast_keys)

    # remove metadata axis name from Variable.sharding
    if spmd.PARTITION_NAME in options.vmap_metadata:
      vectorized_states = [
        spmd.remove_axis(state, index, options.vmap_metadata)
        for state, index in zip(
          vectorized_states, options.variable_axes.values()
        )
      ]

    # merge module state
    module = graphdef.merge(*vectorized_states, broadcast_state)

    output = f(module, *args, **kwargs)

    # split module state
    *vectorized_states_out, broadcast_state_out, moduledef_out = module.split(
      *filters
    )

    # add metadata axis name to Variable.sharding
    if spmd.PARTITION_NAME in options.vmap_metadata:
      vectorized_states_out = [
        spmd.add_axis(state, index, options.vmap_metadata)
        for state, index in zip(
          vectorized_states_out, options.variable_axes.values()
        )
      ]

    return broadcast_state_out, vectorized_states_out, output

  broadcast_state, vectorized_states, output = vmap_fn(
    split_keys, vectorized_states, args, kwargs
  )
  assert moduledef_out is not None

  module.update(((*vectorized_states, broadcast_state), moduledef_out))

  return output


def vmap(
  f: F,
  *,
  variable_axes: tp.Mapping[filterlib.Filter, int] = MappingProxyType({}),
  broadcast_rngs: filterlib.Filter = None,
  in_args_axes: tp.Any = 0,
  in_kwargs_axes: tp.Any = 0,
  out_axes: tp.Any = 0,
  axis_size: int | None = None,
  axis_name: str | None = None,
  spmd_axis_name: str | None = None,
  vmap_metadata: tp.Mapping[str, tp.Any] = MappingProxyType({}),
  is_init: tp.Optional[bool] = None,
) -> F:
  if is_init is None:
    is_init = f.__name__ == '__init__'

  options = VmapOptions(
    variable_axes=variable_axes,
    broadcast_rngs=broadcast_rngs,
    in_args_axes=in_args_axes,
    in_kwargs_axes=in_kwargs_axes,
    out_axes=out_axes,
    axis_size=axis_size,
    axis_name=axis_name,
    spmd_axis_name=spmd_axis_name,
    vmap_metadata=vmap_metadata,
  )

  if is_init:

    @functools.wraps(f)
    def vmap_init_wrapper(module: Module, *args, **kwargs):
      def module_constructor(*args, **kwargs):
        _check_args(args)
        f(module, *args, **kwargs)
        return module

      lifted_module = vmap_init(options, module_constructor, args, kwargs)
      module.update(lifted_module)

    wrapper = vmap_init_wrapper

  else:

    @functools.wraps(f)
    def vmap_apply_wrapper(module: Module, *args, **kwargs) -> tp.Any:
      _check_args(args)
      return vmap_apply(options, f, module, args, kwargs)

    wrapper = vmap_apply_wrapper

  return wrapper  # type: ignore



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""  # Use an empty top-level docstring so Sphinx won't output the one below.
"""Flax error classes.

=== When to create a Flax error class?

If an error message requires more explanation than a one-liner, it is useful to
add it as a separate error class. This may lead to some duplication with
existing documentation or docstrings, but it will provide users with more help
when they are debugging a problem. We can also point to existing documentation
from the error docstring directly.

=== How to name the error class?

* If the error occurs when doing something, name the error
  <Verb><Object><TypeOfError>Error

  For instance, if you want to raise an error when applying a module with an
  invalid method, the error can be: ApplyModuleInvalidMethodError.

 <TypeOfError> is optional, for instance if there is only one error when
  modifying a variable, the error can simply be: ModifyVariableError.

* If there is no concrete action involved the only a description of the error is
  sufficient. For instance: InvalidFilterError, NameInUseError, etc.


=== Copy/pastable template for new error messages:

class Template(FlaxError):
  "" "

  "" "
  def __init__(self):
    super().__init__(f'')
"""


class FlaxError(Exception):
  def __init__(self, message):
    error_page = (
      'https://flax.readthedocs.io/en/latest/api_reference/flax.errors.html'
    )
    module_name = self.__class__.__module__
    class_name = self.__class__.__name__
    error_msg = f'{message} ({error_page}#{module_name}.{class_name})'
    super().__init__(error_msg)


#################################################
# lazy_init.py errors                           #
#################################################


class LazyInitError(FlaxError):
  """Lazy Init function has uncomputable return values.

  This happens when passing an argument to lazy_init with ``jax.ShapeDtypeStruct``
  that affects the initialized variables.
  Make sure the init function only uses the shape and dtype or pass an
  actual JAX array if this is impossible.

  Example::

    class Foo(nn.Module):
      @compact
      def __call__(self, x):
        # This parameter depends on the input x
        # this causes an error when using lazy_init.
        k = self.param("kernel", lambda _: x)
        return x * k
    Foo().lazy_init(random.key(0), jax.ShapeDtypeStruct((8, 4), jnp.float32))
  """

  def __init__(self, partial_val):
    super().__init__(
      'Lazy init encountered a value that could with '
      f'the given inputs (shape: {partial_val}).'
    )


#################################################
# scope.py errors                               #
#################################################


class InvalidRngError(FlaxError):
  """All rngs used in a Module should be passed to

  :meth:`Module.init() <flax.linen.Module.init>` and
  :meth:`Module.apply() <flax.linen.Module.apply>` appropriately. We explain
  both separately using the following example::

    class Bar(nn.Module):
      @nn.compact
      def __call__(self, x):
        some_param = self.param('some_param', nn.initializers.zeros_init(), (1, ))
        dropout_rng = self.make_rng('dropout')
        x = nn.Dense(features=4)(x)
        ...

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        x = Bar()(x)
        ...

  **PRNGs for Module.init()**

  In this example, two rngs are used:

  * ``params`` is used for initializing the parameters of the model. This rng
    is used to initialize the ``some_params`` parameter, and for initializing
    the weights of the ``Dense`` Module used in ``Bar``.

  * ``dropout`` is used for the dropout rng that is used in ``Bar``.

  So, ``Foo`` is initialized as follows::

    init_rngs = {'params': random.key(0), 'dropout': random.key(1)}
    variables = Foo().init(init_rngs, init_inputs)

  If a Module only requires an rng for ``params``, you can use::

       SomeModule().init(rng, ...)  # Shorthand for {'params': rng}


  **PRNGs for Module.apply()**

  When applying ``Foo``, only the rng for ``dropout`` is needed, because
  ``params`` is only used for initializing the Module parameters::

    Foo().apply(variables, inputs, rngs={'dropout': random.key(2)})

  If a Module only requires an rng for ``params``, you don't have to provide
  rngs for apply at all::

       SomeModule().apply(variables, inputs)  # rngs=None
  """

  def __init__(self, msg):
    # For this error message we pass the entire message, since there are various
    # different kinds of RNG errors and we want to be able to be more specific
    # in the error message, while always linking to the same documentation.
    super().__init__(msg)


class ApplyScopeInvalidVariablesTypeError(FlaxError):
  """When calling :meth:`Module.apply() <flax.linen.Module.apply>`, the first

  argument should be a variable dict. For more explanation on variable dicts,
  please see :mod:`flax.core.variables`.
  """

  def __init__(self):
    super().__init__(
      'The first argument passed to an apply function should be '
      'a dictionary of collections. Each collection should be a '
      'dictionary with string keys.'
    )


class ApplyScopeInvalidVariablesStructureError(FlaxError):
  """This error is thrown when the dict passed as ``variables`` to apply() has an

  extra 'params' layer, i.e. {'params': {'params': ...}}.
  For more explanation on variable dicts, please see :mod:`flax.core.variables`.
  """

  def __init__(self, variables):
    super().__init__(
      f'Expect the `variables` (first argument) passed to apply() '
      f'to be a dict with the structure {{"params": ...}}, but got a dict '
      f'with an extra params layer, i.e.  {{"params": {{"params": ... }} }}. '
      f'You should instead pass in your dict\'s ["params"].'
    )


class ScopeParamNotFoundError(FlaxError):
  """This error is thrown when trying to access a parameter that does not exist.

  For instance, in the code below, the initialized embedding name 'embedding'
  does not match the apply name 'embed'::

    class Embed(nn.Module):
      num_embeddings: int
      features: int

      @nn.compact
      def __call__(self, inputs, embed_name='embedding'):
        inputs = inputs.astype('int32')
        embedding = self.param(embed_name,
                               jax.nn.initializers.lecun_normal(),
                               (self.num_embeddings, self.features))
        return embedding[inputs]

    model = Embed(4, 8)
    variables = model.init(random.key(0), jnp.ones((5, 5, 1)))
    _ = model.apply(variables, jnp.ones((5, 5, 1)), 'embed')
  """

  def __init__(self, param_name, scope_path):
    super().__init__(
      f'Could not find parameter named "{param_name}" in scope '
      f'"{scope_path}".'
    )


class ScopeCollectionNotFound(FlaxError):
  """This error is thrown when trying to access a variable from an empty
  collection.

  There are two common causes:

  1. | The collection was not passed to ``apply`` correctly.
     | For example, you might have used ``module.apply(params, ...)`` instead
     | of ``module.apply({'params': params}, ...)``.
  2. | The collection is empty because the variables need to be initialized.
     | In this case, you should have made the collection mutable during
     | apply (e.g.: ``module.apply(variables, ..., mutable=['state'])``.
  """

  def __init__(self, col_name, var_name, scope_path):
    super().__init__(
      f'Tried to access "{var_name}" from collection "{col_name}" in '
      f'"{scope_path}" but the collection is empty.'
    )


class ScopeParamShapeError(FlaxError):
  """This error is thrown when the shape of an existing parameter is different from

  the shape of the return value of the ``init_fn``. This can happen when the
  shape provided during :meth:`Module.apply() <flax.linen.Module.apply>` is
  different from the one used when initializing the module.

  For instance, the following code throws this error because the apply shape
  (``(5, 5, 1)``) is different from the init shape (``(5, 5``). As a result, the
  shape of the kernel during ``init`` is ``(1, 8)``, and the shape during
  ``apply`` is ``(5, 8)``, which results in this error.::

    class NoBiasDense(nn.Module):
      features: int = 8

      @nn.compact
      def __call__(self, x):
        kernel = self.param('kernel',
                            lecun_normal(),
                            (x.shape[-1], self.features))  # <--- ERROR
        y = lax.dot_general(x, kernel,
                            (((x.ndim - 1,), (0,)), ((), ())))
        return y

    variables = NoBiasDense().init(random.key(0), jnp.ones((5, 5, 1)))
    _ = NoBiasDense().apply(variables, jnp.ones((5, 5)))
  """

  def __init__(self, param_name, scope_path, value_shape, init_shape):
    super().__init__(
      f'Initializer expected to generate shape {init_shape} '
      f'but got shape {value_shape} instead for parameter '
      f'"{param_name}" in "{scope_path}".'
    )


class ScopeVariableNotFoundError(FlaxError):
  """This error is thrown when trying to use a variable in a Scope in a collection

  that is immutable. In order to create this variable, mark the collection as
  mutable explicitly using the ``mutable`` keyword in
  :meth:`Module.apply() <flax.linen.Module.apply>`.
  """

  def __init__(self, name, col, scope_path):
    super().__init__(
      f'No Variable named "{name}" for collection "{col}" '
      f'exists in "{scope_path}".'
    )


class InvalidFilterError(FlaxError):
  """A filter should be either a boolean, a string or a container object."""

  def __init__(self, filter_like):
    super().__init__(f'Invalid Filter: "{filter_like}"')


class InvalidScopeError(FlaxError):
  """A temporary Scope is only valid within the context in which it is created::

  with Scope(variables, rngs=rngs).temporary() as root:
    y = fn(root, *args, **kwargs)
    # Here root is valid.
  # Here root is invalid.
  """

  def __init__(self, scope_name):
    super().__init__(f'The scope "{scope_name}" is no longer valid.')


class ModifyScopeVariableError(FlaxError):
  """You cannot update a variable if the collection it belongs to is immutable.

  When you are applying a Module, you should specify which variable collections
  are mutable::

    class MyModule(nn.Module):
      @nn.compact
      def __call__(self, x):
        ...
        var = self.variable('batch_stats', 'mean', ...)
        var.value = ...
        ...

    v = MyModule.init(...)
    ...
    logits = MyModule.apply(v, batch)  # This throws an error.
    logits = MyModule.apply(v, batch, mutable=['batch_stats'])  # This works.
  """

  def __init__(self, col, variable_name, scope_path):
    super().__init__(
      f'Cannot update variable "{variable_name}" in '
      f'"{scope_path}" because collection "{col}" is immutable.'
    )


class JaxTransformError(FlaxError):
  """JAX transforms and Flax modules cannot be mixed.

  JAX's functional transformations expect pure function.
  When you want to use JAX transformations **inside** Flax models,
  you should make use of the Flax transformation wrappers
  (e.g.: ``flax.linen.vmap``, ``flax.linen.scan``, etc.).
  """

  def __init__(self):
    super().__init__('Jax transforms and Flax models cannot be mixed.')


#################################################
# meta.py errors                               #
#################################################


class PartitioningUnspecifiedError(FlaxError):
  """This error is raised when trying to add an axis to a Partitioned variable by

  using a transformation (e.g.: ``scan``, ``vmap``) without specifying the
  "partition_name" in the ``metadata_params`` dict.
  """

  def __init__(self, target):
    super().__init__(
      'Trying to transform a Partitioned variable but "partition_name"'
      f' is not specified in metadata_params: {target}'
    )


#################################################
# module.py errors                              #
#################################################


class NameInUseError(FlaxError):
  """This error is raised when trying to create a submodule, param, or variable

  with an existing name. They are all considered to be in the same namespace.

  **Sharing Submodules**

  This is the wrong pattern for sharing submodules::

    y = nn.Dense(feature=3, name='bar')(x)
    z = nn.Dense(feature=3, name='bar')(x+epsilon)

  Instead, modules should be shared by instance::

    dense = nn.Dense(feature=3, name='bar')
    y = dense(x)
    z = dense(x+epsilon)

  If submodules are not provided with a name, a unique name will be given to
  them automatically::

    class MyModule(nn.Module):
      @nn.compact
      def __call__(self, x):
        x = MySubModule()(x)
        x = MySubModule()(x)  # This is fine.
        return x

  **Parameters and Variables**

  A parameter name can collide with a submodule or variable, since they are all
  stored in the same variable dict::

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        bar = self.param('bar', nn.initializers.zeros_init(), (1, ))
        embed = nn.Embed(num_embeddings=2, features=5, name='bar')  # <-- ERROR!

  Variables should also have unique names, even if they have their own
  collection::

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, inputs):
        _ = self.param('mean', initializers.lecun_normal(), (2, 2))
        _ = self.variable('stats', 'mean', initializers.zeros_init(), (2, 2))
  """

  def __init__(self, key_type, value, module_name):
    # key_type is in {param, variable, submodule}.
    super().__init__(
      f'Could not create {key_type} "{value}" in Module '
      f'{module_name}: Name in use.'
    )


class AssignSubModuleError(FlaxError):
  """You are only allowed to create submodules in two places:

  1.  If your Module is noncompact: inside
      :meth:`Module.setup() <flax.linen.Module.setup>`.
  2.  If your Module is compact: inside the method wrapped in
      :meth:`nn.compact() <flax.linen.compact>`.

  For instance, the following code throws this error, because ``nn.Conv`` is
  created in ``__call__``, which is not marked as compact::

    class Foo(nn.Module):
      def setup(self):
        pass

      def __call__(self, x):
        conv = nn.Conv(features=3, kernel_size=3)

    Foo().init(random.key(0), jnp.zeros((1,)))

  Note that this error is also thrown if you partially defined a Module inside
  setup::

    class Foo(nn.Module):
      def setup(self):
        self.conv = functools.partial(nn.Conv, features=3)

      def __call__(self, x):
        x = self.conv(kernel_size=4)(x)
        return x

    Foo().init(random.key(0), jnp.zeros((1,)))

  In this case, ``self.conv(kernel_size=4)`` is called from ``__call__``, which
  is disallowed because it's neither within ``setup`` nor a method wrapped in
  x``nn.compact``.
  """

  def __init__(self, cls):
    super().__init__(
      f'Submodule {cls} must be defined in `setup()` or in a '
      'method wrapped in `@compact`'
    )


class SetAttributeInModuleSetupError(FlaxError):
  """You are not allowed to modify Module class attributes in

  :meth:`Module.setup() <flax.linen.Module.setup>`::

    class Foo(nn.Module):
      features: int = 6

      def setup(self):
        self.features = 3  # <-- ERROR

      def __call__(self, x):
        return nn.Dense(self.features)(x)

    variables = SomeModule().init(random.key(0), jnp.ones((1, )))

  Instead, these attributes should be set when initializing the Module::

    class Foo(nn.Module):
      features: int = 6

      @nn.compact
      def __call__(self, x):
        return nn.Dense(self.features)(x)

    variables = SomeModule(features=3).init(random.key(0), jnp.ones((1, )))

  TODO(marcvanzee): Link to a design note explaining why it's necessary for
  modules to stay frozen (otherwise we can't safely clone them, which we use for
  lifted transformations).
  """

  def __init__(self):
    super().__init__(f'Module construction attributes are frozen.')


class SetAttributeFrozenModuleError(FlaxError):
  """You can only assign Module attributes to ``self`` inside

  :meth:`Module.setup() <flax.linen.Module.setup>`. Outside of that method, the
  Module instance is frozen (i.e., immutable). This behavior is similar to
  frozen Python dataclasses.

  For instance, this error is raised in the following case::

    class SomeModule(nn.Module):
      @nn.compact
      def __call__(self, x, num_features=10):
        self.num_features = num_features  # <-- ERROR!
        x = nn.Dense(self.num_features)(x)
        return x

    s = SomeModule().init(random.key(0), jnp.ones((5, 5)))

  Similarly, the error is raised when trying to modify a submodule's attributes
  after constructing it, even if this is done in the ``setup()`` method of the
  parent module::

    class Foo(nn.Module):
        def setup(self):
          self.dense = nn.Dense(features=10)
          self.dense.features = 20  # <--- This is not allowed

        def __call__(self, x):
          return self.dense(x)
  """

  def __init__(self, module_cls, attr_name, attr_val):
    super().__init__(
      f"Can't set {attr_name}={attr_val} for Module of type "
      f'{module_cls}: Module instance is frozen outside of '
      'setup method.'
    )


class MultipleMethodsCompactError(FlaxError):
  """The ``@compact`` decorator may only be added to at most one method in a Flax

  module. In order to resolve this, you can:

  * remove ``@compact`` and define submodules and variables using
    :meth:`Module.setup() <flax.linen.Module.setup>`.
  * Use two separate modules that both have a unique ``@compact`` method.

  TODO(marcvanzee): Link to a design note explaining the motivation behind this.
  There is no need for an equivalent to ``hk.transparent`` and it makes submodules
  much more sane because there is no need to prefix the method names.
  """

  def __init__(self):
    super().__init__(f'Only one method per class can be @compact')


class ReservedModuleAttributeError(FlaxError):
  """This error is thrown when creating a Module that is using reserved attributes.

  The following attributes are reserved:

  * ``parent``: The parent Module of this Module.
  * ``name``: The name of this Module.
  """

  def __init__(self, annotations):
    super().__init__(
      f'properties `parent` and `name` are reserved: {annotations}'
    )


class ApplyModuleInvalidMethodError(FlaxError):
  """When calling :meth:`Module.apply() <flax.linen.Module.apply>`, you can specify

  the method to apply using parameter ``method``. This error is thrown if the
  provided parameter is not a method in the Module and not a function with at
  least one argument.

  Learn more on the reference docs for
  :meth:`Module.apply() <flax.linen.Module.apply>`.
  """

  def __init__(self, method):
    super().__init__(
      f'Cannot call apply(): {method} is not a valid function for apply().'
    )


class CallCompactUnboundModuleError(FlaxError):
  """This error occurs when you are trying to call a Module directly, rather than

  through :meth:`Module.apply() <flax.linen.Module.apply>`. For instance, the
  error will be raised when trying to run this code::

    from flax import linen as nn
    import jax.numpy as jnp

    test_dense = nn.Dense(10)
    test_dense(jnp.ones((5,5)))

  Instead, you should pass the variables (parameters and other state) via
  :meth:`Module.apply() <flax.linen.Module.apply>` (or use
  :meth:`Module.init() <flax.linen.Module.init>` to get initial variables)::

    from jax import random
    variables = test_dense.init(random.key(0), jnp.ones((5,5)))

    y = test_dense.apply(variables, jnp.ones((5,5)))
  """

  def __init__(self):
    super().__init__("Can't call compact methods on unbound modules")


class CallSetupUnboundModuleError(FlaxError):
  """This error occurs when you are trying to call ``.setup()`` directly.

  For instance, the error will be raised when trying to run this code::

    from flax import linen as nn
    import jax.numpy as jnp

    class MyModule(nn.Module):
      def setup(self):
        self.submodule = MySubModule()

    module = MyModule()
    module.setup() # <-- ERROR!
    submodule = module.submodule

  In general you shouldn't call ``.setup()`` yourself, if you need to get access
  to a field or submodule defined inside ``setup`` you can instead create a
  function
  to extract it and pass it to ``nn.apply``::

    # setup() will be called automatically by ``nn.apply``
    def get_submodule(module):
      return module.submodule.clone() # avoid leaking the Scope

    empty_variables = {} # you can also use the real variables
    submodule = nn.apply(get_submodule, module)(empty_variables)
  """

  def __init__(self):
    super().__init__("Can't call compact methods on unbound modules")


class CallUnbindOnUnboundModuleError(FlaxError):
  """This error occurs when you are trying to call ``.unbind()`` on an unbound
  Module. For instance, when you try running the following example,
  an error will be raised::

    from flax import linen as nn

    class MyModule(nn.Module):
      @nn.compact
      def __call__(self, x):
        return nn.Dense(features=10)(x)

    module = MyModule()
    module.unbind() # <-- ERROR!

  Instead, you should ``bind`` the Module to a variable collection before calling
  ``.unbind()``::

    bound_module = module.bind(variables)
    ... # do something with bound_module
    module = bound_module.unbind() # <-- OK!
  """

  def __init__(self):
    super().__init__("Can't call `unbind()` on unbound modules")


class InvalidInstanceModuleError(FlaxError):
  """This error occurs when you are trying to call ``.init()``, ``.init_with_output()``, ``.apply()`` or ``.bind()``

  on the Module class itself, instead of an instance of the Module class.
  For example, the error will be raised when trying to run this code::

    class B(nn.Module):
      @nn.compact
      def __call__(self, x):
        return x

    k = random.key(0)
    x = random.uniform(random.key(1), (2,))
    B.init(k, x)   # B is module class, not B() a module instance
    B.apply(vs, x)   # similar issue with apply called on class instead of
    instance.
  """

  def __init__(self):
    super().__init__(
      'Can only call init, init_with_output or apply methods on an instance'
      ' of the Module class, not the Module class itself'
    )


class IncorrectPostInitOverrideError(FlaxError):
  """This error occurs when you overrode ``.__post_init__()`` without calling ``super().__post_init__()``.

  For example, the error will be raised when trying to run this code::

    from flax import linen as nn
    import jax.numpy as jnp
    import jax
    class A(nn.Module):
      x: float
      def __post_init__(self):
        self.x_square = self.x ** 2
        # super().__post_init__() <-- forgot to add this line
      @nn.compact
      def __call__(self, input):
        return input + 3

    r = A(x=3)
    r.init(jax.random.key(2), jnp.ones(3))
  """

  def __init__(self):
    super().__init__(
      'Overrode `.__post_init__()` without calling `super().__post_init__()`'
    )


class DescriptorAttributeError(FlaxError):
  """This error occurs when you are trying to access a property that is accessing a non-existent attribute.

  For example, the error will be raised when trying to run this code::

    class Foo(nn.Module):
        @property
        def prop(self):
            return self.non_existent_field # ERROR!
        def __call__(self, x):
            return self.prop

    foo = Foo()
    variables = foo.init(jax.random.key(0), jnp.ones(shape=(1, 8)))
  """

  def __init__(self):
    super().__init__(
      'Trying to access a property that is accessing a non-existent'
      ' attribute.'
    )


class InvalidCheckpointError(FlaxError):
  """A checkpoint cannot be stored in a directory that already has

  a checkpoint at the current or a later step.

  You can pass ``overwrite=True`` to disable this behavior and
  overwrite existing checkpoints in the target directory.
  """

  def __init__(self, path, step):
    super().__init__(
      f'Trying to save an outdated checkpoint at step: "{step}" and path:'
      f' "{path}".'
    )


class MPACheckpointingRequiredError(FlaxError):
  """To optimally save and restore a multiprocess array (GDA or jax Array outputted from pjit), use GlobalAsyncCheckpointManager.

  You can create an GlobalAsyncCheckpointManager at top-level and pass it as
  argument::

    from jax.experimental.gda_serialization import serialization as gdas
    gda_manager = gdas.GlobalAsyncCheckpointManager()
    save_checkpoint(..., gda_manager=gda_manager)
  """

  def __init__(self, path, step):
    super().__init__(
      f'Checkpoint failed at step: "{step}" and path: "{path}": Target '
      'contains a multiprocess array should be saved/restored with a '
      'GlobalAsyncCheckpointManager.'
    )


class MPARestoreTargetRequiredError(FlaxError):
  """Provide a valid target when restoring a checkpoint with a multiprocess array.

  Multiprocess arrays need a sharding (global meshes and partition specs) to be
  initialized. Therefore, to restore a checkpoint that contains a multiprocess
  array, make sure the ``target`` you passed contains valid multiprocess arrays
  at the corresponding tree structure location. If you cannot provide a full
  valid ``target``, consider ``allow_partial_mpa_restoration=True``.
  """

  def __init__(self, path, step, key=None):
    error_msg = (
      f'Restore checkpoint failed at step: "{step}" and path: "{path}": '
      'Checkpoints containing a multiprocess array need to be restored with '
      'a target with pre-created arrays. If you cannot provide a full valid '
      'target, consider ``allow_partial_mpa_restoration=True``. '
    )
    if key:
      error_msg += f'This error fired when trying to restore array at {key}.'
    super().__init__(error_msg)


class MPARestoreDataCorruptedError(FlaxError):
  """A multiprocess array stored in Google Cloud Storage doesn't contain a "commit_success.txt" file, which should be written at the end of the save.

  Failure of finding it could indicate a corruption of your saved GDA data.
  """

  def __init__(self, step, path):
    super().__init__(
      f'Restore checkpoint failed at step: "{step}" on multiprocess array at '
      f' "{path}": No "commit_success.txt" found on this "_gda" directory. '
      'Was its save halted before completion?'
    )


#################################################
# transforms.py errors                          #
#################################################


class TransformedMethodReturnValueError(FlaxError):
  """Transformed Module methods cannot return other Modules or Variables."""

  def __init__(self, name):
    super().__init__(
      f'Transformed module method {name} cannot return Modules or Variables.'
    )


class TransformTargetError(FlaxError):
  """Linen transformations must be applied to Modules classes or functions taking a Module instance as the first argument.

  This error occurs when passing an invalid target to a linen transform
  (nn.vmap, nn.scan, etc.). This occurs for example when trying to
  transform a Module instance::

    nn.vmap(nn.Dense(features))(x)  # raises TransformTargetError

  You can transform the ``nn.Dense`` class directly instead::

    nn.vmap(nn.Dense)(features)(x)

  Or you can create a function that takes the module instance as the first
  argument::

    class BatchDense(nn.Module):
      @nn.compact
      def __call__(self, x):
        return nn.vmap(
            lambda mdl, x: mdl(x),
            variable_axes={'params': 0}, split_rngs={'params':
            True})(nn.Dense(3), x)
  """

  def __init__(self, target):
    super().__init__(
      'Linen transformations must be applied to Modules classes or'
      ' functions taking a Module instance as the first argument.'
      f' The provided target is not a Module class or callable: {target}'
    )


#################################################
# io.py errors                                  #
#################################################


class AlreadyExistsError(FlaxError):
  """Attempting to overwrite a file via copy.

  You can pass ``overwrite=True`` to disable this behavior and overwrite
  existing files in.
  """

  def __init__(self, path):
    super().__init__(f'Trying overwrite an existing file: "{path}".')


#################################################
# cursor.py errors                                  #
#################################################


class CursorFindError(FlaxError):
  """Error when calling :meth:`Cursor.find() <flax.cursor.Cursor.find>`.

  This error occurs if no object or more than one object is found, given
  the conditions of the ``cond_fn``.
  """

  def __init__(self, cursor=None, cursor2=None):
    if cursor and cursor2:
      super().__init__(
        'More than one object found given the conditions of the cond_fn. '
        'The first two objects found have the following paths: '
        f'{cursor._path} and {cursor2._path}'
      )
    else:
      super().__init__('No object found given the conditions of the cond_fn.')


class TraverseTreeError(FlaxError):
  """Error when calling ``Cursor._traverse_tree()``. This function has two
  modes:

  - if ``update_fn`` is not None, it will traverse the tree and return a
    generator of tuples containing the path where the ``update_fn`` was
    applied and the newly modified value.
  - if ``cond_fn`` is not None, it will traverse the tree and return a
    generator of tuple paths that fulfilled the conditions of the ``cond_fn``.

  This error occurs if either both ``update_fn`` and ``cond_fn`` are None,
  or both are not None.
  """

  def __init__(self, update_fn, cond_fn):
    if update_fn is None and cond_fn is None:
      super().__init__(
        'Both update_fn and cond_fn are None. Exactly one of them must be'
        ' None.'
      )
    else:
      super().__init__(
        'Both update_fn and cond_fn are not None. Exactly one of them must be'
        ' not None.'
      )



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.linen.batch_apply."""

import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized

from flax import linen as nn

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class BatchApplyTest(parameterized.TestCase):
  @parameterized.parameters(
    {'fn': lambda a, b: a + b.reshape(1, -1)},
    {'fn': lambda a, b: jnp.dot(a, b)},
  )
  def test_batchapply(self, fn):
    a = jax.random.normal(jax.random.key(0), [2, 3, 4])
    b = jax.random.normal(jax.random.key(1), [4])

    def raises(a, b):
      if len(a.shape) != 2:
        raise ValueError('a must be shape 2')
      if len(b.shape) != 1:
        raise ValueError('b must be shape 1')
      return fn(a, b)

    out = nn.BatchApply(raises)(a, b)
    expected_merged_leading = raises(a.reshape(2 * 3, 4), b)
    expected = expected_merged_leading.reshape(
      (2, 3) + expected_merged_leading.shape[1:]
    )
    np.testing.assert_array_equal(out, expected)

  def test_batchapply_accepts_float(self):
    def raises(a, b):
      if len(a.shape) != 2:
        raise ValueError('a must be shape 2')
      return a + b

    out = nn.BatchApply(raises)(jnp.ones([2, 3, 4]), 2.0)
    np.testing.assert_array_equal(out, 3 * jnp.ones([2, 3, 4]))

  def test_batchapply_accepts_none(self):
    def raises(a, b):
      if a is not None:
        raise ValueError('a must be None.')
      if len(b.shape) != 2:
        raise ValueError('b must be shape 2')
      return 3 * b

    out = nn.BatchApply(raises)(None, jnp.ones([2, 3, 4]))
    np.testing.assert_array_equal(out, 3 * jnp.ones([2, 3, 4]))

  def test_batchapply_raises(self):
    with self.assertRaisesRegex(ValueError, 'requires at least one input'):
      nn.BatchApply(lambda: 1)()


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax API."""

# pylint: disable=g-import-not-at-top
# pyformat: disable

from flax import configurations
config: configurations.Config = configurations.config
del configurations

from flax import core
from flax import jax_utils
from flax import linen
from flax import serialization
from flax import traverse_util

from flax import version
__version__: str = version.__version__
del version

# DO NOT REMOVE - Marker for internal deprecated API.

# DO NOT REMOVE - Marker for internal logging.

# pyformat: enable
# pylint: enable=g-import-not-at-top



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

import jax

from flax.experimental import nnx

A = tp.TypeVar('A')


class TestVariable:
  def test_value(self):
    r1 = nnx.Variable(1)
    assert r1.value == 1

    r2 = jax.tree_map(lambda x: x + 1, r1)

    assert r1.value == 1
    assert r2.value == 2
    assert r1 is not r2



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from typing import List

import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest
from jax import random

from flax import linen as nn
from flax import struct
from flax.core.scope import Array
from flax.linen import summary

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()

CONSOLE_TEST_KWARGS = dict(force_terminal=False, no_color=True, width=10_000)


def _get_shapes(pytree):
  return jax.tree_util.tree_map(
    lambda x: x.shape if hasattr(x, 'shape') else x, pytree
  )


def _get_obj_repr_value(x):
  if isinstance(x, summary._ObjectRepresentation):
    return x.obj
  return x


class ConvBlock(nn.Module):
  features: int
  kernel_size: List[int]
  test_sow: bool

  def setup(self) -> None:
    self.conv = nn.Conv(self.features, self.kernel_size)
    self.bn = nn.BatchNorm()
    self.dropout = nn.Dropout(0.5)

  def block_method(self, x: Array, training: bool) -> Array:
    x = self.conv(x)

    if self.test_sow:
      self.sow('intermediates', 'INTERM', x)

    x = self.bn(x, use_running_average=not training)
    x = self.dropout(x, deterministic=not training)
    x = nn.relu(x)
    return x

  def __call__(self, x: Array, training: bool) -> Array:
    x = self.conv(x)

    if self.test_sow:
      self.sow('intermediates', 'INTERM', x)

    x = self.bn(x, use_running_average=not training)
    x = self.dropout(x, deterministic=not training)
    x = nn.relu(x)
    return x


class CNN(nn.Module):
  test_sow: bool

  def setup(self) -> None:
    self.block1 = ConvBlock(32, [3, 3], test_sow=self.test_sow)
    self.block2 = ConvBlock(64, [3, 3], test_sow=self.test_sow)
    self.dense = nn.Dense(10)

  def cnn_method(self, x: Array, training: bool) -> Array:
    x = self.block1.block_method(x, training=training)
    x = self.block2.block_method(x, training=training)
    x = x.mean(axis=(1, 2))

    if self.test_sow:
      self.sow('intermediates', 'INTERM', x)

    x = self.dense(x)

    return x, dict(a=x, b=x + 1.0)

  def __call__(self, x: Array, training: bool) -> Array:
    x = self.block1.block_method(x, training=training)
    x = self.block2.block_method(x, training=training)
    x = x.mean(axis=(1, 2))

    if self.test_sow:
      self.sow('intermediates', 'INTERM', x)

    x = self.dense(x)

    return x, dict(a=x, b=x + 1.0)


class SummaryTest(absltest.TestCase):
  def test_module_summary(self):
    """
    This test creates a Table using `module_summary` and checks that it
    matches the expected output given the CNN model defined in
    `_get_tabulate_cnn`.
    """

    batch_size = 32

    x = jnp.ones((batch_size, 28, 28, 1))
    module = CNN(test_sow=False)

    table = summary._get_module_table(
      module,
      depth=None,
      show_repeated=True,
      compute_flops=True,
      compute_vjp_flops=True,
    )(
      {'dropout': random.key(0), 'params': random.key(1)},
      x,
      training=True,
      mutable=True,
    )
    # get values for inputs and outputs from their _ValueRepresentation
    for row in table:
      row.inputs = jax.tree_util.tree_map(_get_obj_repr_value, row.inputs)
      row.outputs = jax.tree_util.tree_map(_get_obj_repr_value, row.outputs)

    # 10 rows = 1 CNN + 4 ConvBlock_0 + 4 ConvBlock_1 + 1 Dense_0
    self.assertLen(table, 10)

    # check paths
    self.assertEqual(table[0].path, ())

    self.assertEqual(table[1].path, ('block1',))
    self.assertEqual(table[2].path, ('block1', 'conv'))
    self.assertEqual(table[3].path, ('block1', 'bn'))
    self.assertEqual(table[4].path, ('block1', 'dropout'))

    self.assertEqual(table[5].path, ('block2',))
    self.assertEqual(table[6].path, ('block2', 'conv'))
    self.assertEqual(table[7].path, ('block2', 'bn'))
    self.assertEqual(table[8].path, ('block2', 'dropout'))

    self.assertEqual(table[9].path, ('dense',))

    # check outputs shapes
    self.assertEqual(
      (table[0].inputs[0].shape, table[0].inputs[1]),
      (x.shape, dict(training=True)),
    )
    self.assertEqual(
      _get_shapes(table[0].outputs),
      ((batch_size, 10), dict(a=(batch_size, 10), b=(batch_size, 10))),
    )

    self.assertEqual(
      _get_shapes(table[1].inputs),
      ((batch_size, 28, 28, 1), {'training': True}),
    )
    self.assertEqual(table[1].outputs.shape, (batch_size, 28, 28, 32))
    self.assertEqual(table[2].inputs.shape, (batch_size, 28, 28, 1))
    self.assertEqual(table[2].outputs.shape, (batch_size, 28, 28, 32))
    self.assertEqual(
      _get_shapes(table[3].inputs),
      ((batch_size, 28, 28, 32), {'use_running_average': False}),
    )
    self.assertEqual(table[3].outputs.shape, (batch_size, 28, 28, 32))
    self.assertEqual(
      _get_shapes(table[4].inputs),
      ((batch_size, 28, 28, 32), {'deterministic': False}),
    )
    self.assertEqual(table[4].outputs.shape, (batch_size, 28, 28, 32))

    self.assertEqual(
      _get_shapes(table[5].inputs),
      ((batch_size, 28, 28, 32), {'training': True}),
    )
    self.assertEqual(table[5].outputs.shape, (batch_size, 28, 28, 64))
    self.assertEqual(table[6].inputs.shape, (batch_size, 28, 28, 32))
    self.assertEqual(table[6].outputs.shape, (batch_size, 28, 28, 64))
    self.assertEqual(
      _get_shapes(table[7].inputs),
      ((batch_size, 28, 28, 64), {'use_running_average': False}),
    )
    self.assertEqual(table[7].outputs.shape, (batch_size, 28, 28, 64))
    self.assertEqual(
      _get_shapes(table[8].inputs),
      ((batch_size, 28, 28, 64), {'deterministic': False}),
    )
    self.assertEqual(table[8].outputs.shape, (batch_size, 28, 28, 64))

    self.assertEqual(table[9].inputs.shape, (batch_size, 64))
    self.assertEqual(table[9].outputs.shape, (batch_size, 10))

    # check no summary is performed
    for row in table:
      self.assertEqual(
        row.module_variables,
        row.counted_variables,
      )

    # Each module FLOPs >= sum of its submodule FLOPs.
    # Can be greater due to ops like `nn.relu` not belonging to any submodule.
    for r in table:
      flops, vjp_flops = r.flops, r.vjp_flops
      submodule_flops, submodule_vjp_flops = 0, 0
      for s in table:
        if len(s.path) == len(r.path) + 1 and s.path[: len(r.path)] == r.path:
          submodule_flops += s.flops
          submodule_vjp_flops += s.vjp_flops

      self.assertGreaterEqual(flops, submodule_flops)
      self.assertGreaterEqual(vjp_flops, submodule_vjp_flops)

  def test_module_summary_with_depth(self):
    """
    This test creates a Table using `module_summary` set the `depth` argument
    to `1`, table should have fewer rows as a consequence.
    """
    batch_size = 32

    x = jnp.ones((batch_size, 28, 28, 1))
    module = CNN(test_sow=False)

    table = summary._get_module_table(
      module,
      depth=1,
      show_repeated=True,
      compute_flops=True,
      compute_vjp_flops=True,
    )(
      {'dropout': random.key(0), 'params': random.key(1)},
      x,
      training=True,
      mutable=True,
    )
    # get values for inputs and outputs from their _ValueRepresentation

    for row in table:
      row.inputs = jax.tree_util.tree_map(_get_obj_repr_value, row.inputs)
      row.outputs = jax.tree_util.tree_map(_get_obj_repr_value, row.outputs)

    # 4 rows = 1 CNN + 1 ConvBlock_0 + 1 ConvBlock_1 + 1 Dense_0
    self.assertLen(table, 4)

    # check paths
    self.assertEqual(table[0].path, ())

    self.assertEqual(table[1].path, ('block1',))
    self.assertEqual(table[2].path, ('block2',))
    self.assertEqual(table[3].path, ('dense',))

    # check outputs shapes
    self.assertEqual(
      (table[0].inputs[0].shape, table[0].inputs[1]),
      (x.shape, dict(training=True)),
    )
    self.assertEqual(
      _get_shapes(table[0].outputs),
      ((batch_size, 10), dict(a=(batch_size, 10), b=(batch_size, 10))),
    )

    self.assertEqual(
      _get_shapes(table[1].inputs),
      ((batch_size, 28, 28, 1), {'training': True}),
    )
    self.assertEqual(table[1].outputs.shape, (batch_size, 28, 28, 32))

    self.assertEqual(
      _get_shapes(table[2].inputs),
      ((batch_size, 28, 28, 32), {'training': True}),
    )
    self.assertEqual(table[2].outputs.shape, (batch_size, 28, 28, 64))

    self.assertEqual(table[3].inputs.shape, (batch_size, 64))
    self.assertEqual(table[3].outputs.shape, (batch_size, 10))

    # check ConvBlock_0 and ConvBlock_1 are summarized
    self.assertNotEqual(table[1].module_variables, table[1].counted_variables)
    self.assertNotEqual(table[2].module_variables, table[2].counted_variables)

    # check CNN and Dense_0 output are not summarized
    self.assertEqual(table[0].module_variables, table[0].counted_variables)
    self.assertEqual(table[3].module_variables, table[3].counted_variables)

    # Top level FLOPs > sum of listed submodule FLOPs, since not all are listed.
    self.assertGreater(table[0].flops, sum(r.flops for r in table[1:]))
    self.assertGreater(table[0].vjp_flops, sum(r.vjp_flops for r in table[1:]))

  def test_tabulate(self):
    """
    This test creates a string representation of a Module using `Module.tabulate`
    and checks that it matches the expected output given the CNN model defined
    in `_get_tabulate_cnn`.
    """
    batch_size = 32

    x = jnp.ones((batch_size, 28, 28, 1))
    module = CNN(test_sow=False)

    module_repr = module.tabulate(
      {'dropout': random.key(0), 'params': random.key(1)},
      x,
      training=True,
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )

    # NOTE: it's tricky to validate the content of lines
    # because it seems to be shell-dependent, so we will
    # just check lines that won't change between environments
    lines = module_repr.split('\n')

    # check title
    module_name = module.__class__.__name__
    self.assertIn(f'{module_name} Summary', lines[1])

    # check headers are correct
    self.assertIn('path', lines[3])
    self.assertIn('module', lines[3])
    self.assertIn('inputs', lines[3])
    self.assertIn('outputs', lines[3])
    self.assertIn('params', lines[3])
    self.assertIn('flops', lines[3])
    self.assertIn('vjp_flops', lines[3])
    self.assertIn('batch_stats', lines[3])

    # collection counts
    self.assertIn('Total', lines[-6])
    self.assertIn('192', lines[-6])
    self.assertIn('768 B', lines[-6])
    self.assertIn('19,658', lines[-6])
    self.assertIn('78.6 KB', lines[-6])

    # total counts
    self.assertIn('Total Parameters', lines[-3])
    self.assertIn('19,850', lines[-3])
    self.assertIn('79.4 KB', lines[-3])

  def test_tabulate_with_sow(self):
    batch_size = 32

    x = jnp.ones((batch_size, 28, 28, 1))
    module = CNN(test_sow=True)

    module_repr = module.tabulate(
      {'dropout': random.key(0), 'params': random.key(1)},
      x,
      training=True,
      mutable=True,
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )

    self.assertIn('intermediates', module_repr)
    self.assertIn('INTERM', module_repr)

  def test_tabulate_with_method(self):
    batch_size = 32

    x = jnp.ones((batch_size, 28, 28, 1))
    module = CNN(test_sow=False)

    module_repr = module.tabulate(
      {'dropout': random.key(0), 'params': random.key(1)},
      x,
      training=True,
      method=CNN.cnn_method,
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )

    self.assertIn('(block_method)', module_repr)
    self.assertIn('(cnn_method)', module_repr)

  def test_tabulate_function(self):
    """
    This test creates a string representation of a Module using
    `Module.tabulate` and checks that it matches the expected output given the
    CNN model defined in `_get_tabulate_cnn`.
    """
    batch_size = 32

    x = jnp.ones((batch_size, 28, 28, 1))
    module = CNN(test_sow=False)

    module_repr = nn.tabulate(
      module,
      {'dropout': random.key(0), 'params': random.key(1)},
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )(x, training=True)

    lines = module_repr.split('\n')

    # check title
    module_name = module.__class__.__name__
    self.assertIn(f'{module_name} Summary', lines[1])

    # check headers are correct
    self.assertIn('path', lines[3])
    self.assertIn('module', lines[3])
    self.assertIn('inputs', lines[3])
    self.assertIn('outputs', lines[3])
    self.assertIn('params', lines[3])
    self.assertIn('flops', lines[3])
    self.assertIn('batch_stats', lines[3])

    # collection counts
    self.assertIn('Total', lines[-6])
    self.assertIn('192', lines[-6])
    self.assertIn('768 B', lines[-6])
    self.assertIn('19,658', lines[-6])
    self.assertIn('78.6 KB', lines[-6])

    # total counts
    self.assertIn('Total Parameters', lines[-3])
    self.assertIn('19,850', lines[-3])
    self.assertIn('79.4 KB', lines[-3])

  def test_lifted_transform(self):
    class LSTM(nn.Module):
      features: int

      @nn.compact
      def __call__(self, x):
        carry = nn.LSTMCell(self.features).initialize_carry(
          random.key(0), x[:, 0].shape
        )
        ScanLSTM = nn.scan(
          nn.LSTMCell,
          variable_broadcast='params',
          split_rngs={'params': False},
          in_axes=1,
          out_axes=1,
        )
        return ScanLSTM(self.features, name='ScanLSTM')(carry, x)

    lstm = LSTM(features=128)

    with jax.check_tracer_leaks(True):
      module_repr = lstm.tabulate(
        random.key(0),
        x=jnp.ones((32, 128, 64)),
        console_kwargs=CONSOLE_TEST_KWARGS,
        compute_flops=True,
        compute_vjp_flops=True,
      )

    lines = module_repr.splitlines()

    self.assertIn('LSTM', lines[5])
    self.assertIn('ScanLSTM', lines[9])
    self.assertIn('LSTMCell', lines[9])
    self.assertIn('ScanLSTM/ii', lines[13])
    self.assertIn('Dense', lines[13])

  def test_lifted_transform_no_rename(self):
    class LSTM(nn.Module):
      features: int

      @nn.compact
      def __call__(self, x):
        carry = nn.LSTMCell(self.features).initialize_carry(
          random.key(0), x[:, 0].shape
        )
        ScanLSTM = nn.scan(
          nn.LSTMCell,
          variable_broadcast='params',
          split_rngs={'params': False},
          in_axes=1,
          out_axes=1,
        )
        return ScanLSTM(self.features)(carry, x)

    lstm = LSTM(features=128)

    with jax.check_tracer_leaks(True):
      module_repr = lstm.tabulate(
        random.key(0),
        x=jnp.ones((32, 128, 64)),
        console_kwargs=CONSOLE_TEST_KWARGS,
        compute_flops=True,
        compute_vjp_flops=True,
      )

    lines = module_repr.splitlines()

    self.assertIn('LSTM', lines[5])
    self.assertIn('ScanLSTMCell_0', lines[9])
    self.assertIn('LSTMCell', lines[9])
    self.assertIn('ScanLSTMCell_0/ii', lines[13])
    self.assertIn('Dense', lines[13])

  def test_module_reuse(self):
    class ConvBlock(nn.Module):
      @nn.compact
      def __call__(self, x):
        x = nn.Conv(32, [3, 3])(x)
        x = nn.BatchNorm(use_running_average=True)(x)
        x = nn.Dropout(0.5, deterministic=True)(x)
        x = nn.relu(x)
        return x

    class CNN(nn.Module):
      @nn.compact
      def __call__(self, x):
        block = ConvBlock()
        x = block(x)
        x = block(x)
        x = block(x)
        return x

    x = jnp.ones((4, 28, 28, 32))
    module_repr = CNN().tabulate(
      jax.random.key(0),
      x=x,
      show_repeated=True,
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )
    lines = module_repr.splitlines()

    # first call
    self.assertIn('ConvBlock_0/Conv_0', lines[9])
    self.assertIn('bias', lines[9])
    self.assertIn('ConvBlock_0/BatchNorm_0', lines[14])
    self.assertIn('mean', lines[14])
    self.assertIn('bias', lines[14])
    self.assertIn('ConvBlock_0/Dropout_0', lines[19])

    # second call
    self.assertIn('ConvBlock_0/Conv_0', lines[23])
    self.assertNotIn('bias', lines[23])
    self.assertIn('ConvBlock_0/BatchNorm_0', lines[25])
    self.assertNotIn('mean', lines[25])
    self.assertNotIn('bias', lines[25])
    self.assertIn('ConvBlock_0/Dropout_0', lines[27])

    # third call
    self.assertIn('ConvBlock_0/Conv_0', lines[31])
    self.assertNotIn('bias', lines[31])
    self.assertIn('ConvBlock_0/BatchNorm_0', lines[33])
    self.assertNotIn('mean', lines[33])
    self.assertNotIn('bias', lines[33])
    self.assertIn('ConvBlock_0/Dropout_0', lines[35])

    # Test that CNN FLOPs are 3x ConvBlock FLOPs.
    args = ({'dropout': random.key(0), 'params': random.key(1)}, x)
    cnn = summary._get_module_table(
      CNN(),
      depth=1,
      show_repeated=True,
      compute_flops=True,
      compute_vjp_flops=True,
    )(*args, mutable=True)

    block = summary._get_module_table(
      ConvBlock(),
      depth=1,
      show_repeated=True,
      compute_flops=True,
      compute_vjp_flops=True,
    )(*args, mutable=True)

    # Total forward/backward FLOPs equal to their sums of sub blocks.
    self.assertEqual(cnn[0].flops, sum(r.flops for r in cnn[1:]))
    self.assertEqual(cnn[0].vjp_flops, sum(r.vjp_flops for r in cnn[1:]))

    # Each sub block has cost equal to ConvBlock instantiated separately.
    for r in cnn[1:]:
      self.assertEqual(r.flops, block[0].flops)
      self.assertEqual(r.vjp_flops, block[0].vjp_flops)

  def test_empty_input(self):
    class EmptyInput(nn.Module):
      @nn.compact
      def __call__(self):
        return 1

    module = EmptyInput()
    module_repr = module.tabulate(
      {},
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )
    lines = module_repr.splitlines()

    # 1 output and 0 forward / backward FLOPs.
    self.assertRegex(
      lines[5], r'│\s*│\s*EmptyInput\s*│\s*│\s*1\s*│\s*0\s*│\s*0\s*│'
    )

  def test_numpy_scalar(self):
    class Submodule(nn.Module):
      def __call__(self, x):
        return x + 1

    class EmptyInput(nn.Module):
      @nn.compact
      def __call__(self):
        return Submodule()(x=np.pi)

    module = EmptyInput()
    module_repr = module.tabulate(
      {},
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )
    lines = module_repr.splitlines()

    self.assertIn('4.141592', lines[5])
    self.assertIn('x: 3.141592', lines[7])
    self.assertIn('4.141592', lines[7])

    # 0 forward / backward FLOPs due to precomputed output values.
    self.assertIn('│ 0     │ 0', lines[5])
    self.assertIn('│ 0     │ 0', lines[7])

  def test_partitioned_params(self):
    class Classifier(nn.Module):
      @nn.compact
      def __call__(self, x):
        hidden = nn.Dense(
          features=1024,
          kernel_init=nn.with_partitioning(
            nn.initializers.lecun_normal(), (None, 'data')
          ),
          bias_init=nn.with_partitioning(nn.initializers.zeros, (None,)),
          name='hidden',
        )
        x = x / 255.0
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.relu(hidden(x))
        x = nn.Dense(features=10, name='head')(x)
        return x

    module = Classifier()
    lines = module.tabulate(
      jax.random.key(0),
      jnp.empty((1, 28, 28, 1)),
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    ).splitlines()
    self.assertIn('P(None,)', lines[7])
    self.assertIn('P(None, data)', lines[8])

    # Per-layer forward FLOPs:
    self.assertIn('1606656', lines[7])  # 1 * (28 * 28 * 1) * 1024 * 2 + 1024
    self.assertIn('20490', lines[12])  #  1 * (1024       ) * 10   * 2 + 10

    # Total forward FLOPs: input division + ReLU + two dense layers above.
    # (1 * 28 * 28 * 1) + (1 * 1024) + 1606656 + 20490.
    self.assertIn('1628954', lines[5])

    # Per-layer backward FLOPs:

    # [3x MMs: forward, input cotangent, weight cotangent] 1024 * 784 * 2 * 3
    # + [forward bias addition] 1024
    # + [`mutable=True`: weight and bias sizes] 1024 * 784 + 1024
    self.assertIn('5621760', lines[7])

    # [3x matmuls: forward, input cotangent, weight cotangent] 1024 * 10 * 2 * 3
    # + [forward bias addition] 10
    # + [`mutable=True`: weight and bias sizes] 1024 * 10 + 10
    self.assertIn('71700', lines[12])

    # Total backward FLOPs: input division + ReLU + two dense layers above.
    # 2 * (1 * 28 * 28 * 1) + 3 * (1 * 1024) + 5621760 + 71700.
    self.assertIn('5698100', lines[5])

  def test_non_array_variables(self):
    class Metadata(struct.PyTreeNode):
      names: tuple = struct.field(pytree_node=False)

    class Foo(nn.Module):
      @nn.compact
      def __call__(self):
        self.sow('foo', 'bar', Metadata(('baz', 'qux')))

    module = Foo()
    lines = module.tabulate(
      {},
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    ).splitlines()
    self.assertIn('names', lines[6])
    self.assertIn('baz', lines[7])
    self.assertIn('qux', lines[8])

    # 0 forward and backward FLOPs.
    self.assertIn('│ 0     │ 0', lines[5])

  def test_tabulate_param_count_and_flops(self):
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        h = nn.Dense(4)(x)
        return nn.Dense(2)(h)

    module = Foo()
    rng = jax.random.key(0)
    x = jnp.ones((16, 9))

    rep = module.tabulate(
      rng,
      x,
      console_kwargs=CONSOLE_TEST_KWARGS,
      compute_flops=True,
      compute_vjp_flops=True,
    )
    lines = rep.splitlines()
    self.assertIn('Total Parameters: 50', lines[-2])

  def test_tabulate_enum(self):
    class Net(nn.Module):
      @nn.compact
      def __call__(self, inputs):
        x = inputs['x']
        x = nn.Dense(features=2)(x)
        return jnp.sum(x)

    class InputEnum(str, enum.Enum):
      x = 'x'

    inputs = {InputEnum.x: jnp.ones((1, 1))}
    # test args
    lines = Net().tabulate(jax.random.key(0), inputs).split('\n')
    self.assertIn('x: \x1b[2mfloat32\x1b[0m[1,1]', lines[5])
    # test kwargs
    lines = Net().tabulate(jax.random.key(0), inputs=inputs).split('\n')
    self.assertIn('inputs:', lines[5])
    self.assertIn('x: \x1b[2mfloat32\x1b[0m[1,1]', lines[6])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for constructing an iterator which prefetches data asynchronously."""

import threading
import warnings


class PrefetchIterator:
  """Wraps an iterator to provide async prefetching.

  DEPRECATION WARNING:
  TensorFlow datasets no longer require manual prefetching.

  Previously this class was used to make data loading using TensorFlow datasets
  more efficient. Now TF data handles prefetching with NumPy iterators
  correctly.

  Example::

    tf_iter = dataset.as_numpy_iterator()  # only loads data while calling next
    tf_iter = PrefetchIterator(tf_iter)  # prefetches data in the background

  """

  def __init__(self, data_iter, buffer_size=1):
    """Construct a PrefetchIterator.

    Args:
      data_iter: the Iterator that should be prefetched.
      buffer_size: how many items to prefetch (default: 1).
    """
    warnings.warn(
      'PrefetchIterator is deprecated. Use the standard `tf.data`'
      ' prefetch method instead',
      DeprecationWarning,
    )

    self._data_iter = data_iter
    self.buffer_size = buffer_size
    self._cond = threading.Condition()
    self._buffer = []
    self._active = True
    self._thread = threading.Thread(target=self._prefetch_loop, daemon=True)
    self._thread.start()
    self._error = None

  def __iter__(self):
    return self

  def __next__(self):
    with self._cond:
      self._cond.wait_for(lambda: self._buffer or not self._active)
      if self._buffer:
        item = self._buffer.pop(0)
        self._cond.notify_all()
        return item
      if self._error:
        raise self._error  # pylint: disable=raising-bad-type
      assert not self._active
      raise StopIteration()

  def close(self):
    with self._cond:
      self._active = False
      self._cond.notify_all()

  def _prefetch_loop(self):
    """Prefetch loop that prefetches a tf dataset."""

    def _predicate():
      return len(self._buffer) < self.buffer_size or not self._active

    while True:
      try:
        item = next(self._data_iter)
        with self._cond:
          self._buffer.append(item)
          self._cond.notify_all()
          self._cond.wait_for(_predicate)
          if not self._active:
            return
      except Exception as e:  # pylint: disable=broad-except
        with self._cond:
          self._error = e
          self._active = False
          self._cond.notify_all()
          return



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.metrics.tensorboard."""
import itertools
import pathlib
import tempfile

import numpy as np
import tensorflow as tf
from absl.testing import absltest
from tensorboard.backend.event_processing import (
  directory_watcher,
  event_file_loader,
)
from tensorboard.util import tensor_util

from flax.metrics.tensorboard import SummaryWriter, _flatten_dict


def _process_event(event):
  for value in event.summary.value:
    yield {'wall_time': event.wall_time, 'step': event.step, 'value': value}


def _disk_usage(path: pathlib.Path):
  """Recursively computes the disk usage of a directory."""
  if path.is_file():
    return path.stat().st_size
  elif path.is_dir():
    size_bytes = 0
    for file in path.iterdir():
      size_bytes += _disk_usage(file)
    return size_bytes
  else:
    raise NotImplementedError('What filetype is {file}?')


class TensorboardTest(absltest.TestCase):
  def parse_and_return_summary_value(self, path):
    """Parse the event file in the given path and return the
    only summary value."""
    event_value_list = []
    event_file_generator = directory_watcher.DirectoryWatcher(
      path, event_file_loader.EventFileLoader
    ).Load()
    event_values = itertools.chain.from_iterable(
      map(_process_event, event_file_generator)
    )
    for value_dict in event_values:
      event_value_list.append(value_dict)

    self.assertLen(event_value_list, 1)
    self.assertEqual(event_value_list[0]['step'], 1)
    self.assertGreater(event_value_list[0]['wall_time'], 0.0)
    return event_value_list[0]['value']

  def test_summarywriter_flush_after_close(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    summary_writer.close()
    with self.assertRaises(AttributeError):
      summary_writer.flush()

  def test_summarywriter_scalar(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    # Write the scalar and check if the event exists and check data.
    float_value = 99.1232
    summary_writer.scalar(tag='scalar_test', value=float_value, step=1)

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'scalar_test')
    self.assertTrue(
      np.allclose(
        tensor_util.make_ndarray(summary_value.tensor).item(), float_value
      )
    )

  def test_summarywriter_text(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    text = 'hello world.'
    summary_writer.text(tag='text_test', textdata=text, step=1)

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'text_test')
    self.assertEqual(
      tensor_util.make_ndarray(summary_value.tensor).item().decode('utf-8'),
      text,
    )

  def test_summarywriter_image(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    expected_img = np.random.uniform(low=0.0, high=255.0, size=(30, 30, 3))
    expected_img = expected_img.astype(np.uint8)
    summary_writer.image(tag='image_test', image=expected_img, step=1)
    summary_value = self.parse_and_return_summary_value(path=log_dir)

    self.assertEqual(summary_value.tag, 'image_test')
    actual_img = tf.image.decode_image(summary_value.tensor.string_val[2])
    self.assertTrue(np.allclose(actual_img, expected_img))

  def test_summarywriter_image_float_pixel_values(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    expected_img = np.random.uniform(low=0.0, high=1.0, size=(30, 30, 3))
    summary_writer.image(tag='image_test', image=expected_img, step=1)
    summary_value = self.parse_and_return_summary_value(path=log_dir)

    # convert and scale expected_img appropriately to numpy uint8.
    expected_img = tf.image.convert_image_dtype(
      image=expected_img, dtype=np.uint8
    )

    self.assertEqual(summary_value.tag, 'image_test')
    actual_img = tf.image.decode_image(summary_value.tensor.string_val[2])
    self.assertTrue(np.allclose(actual_img, expected_img))

  def test_summarywriter_2dimage_scaled(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    img = np.random.uniform(low=0.0, high=255.0, size=(30, 30))
    img = img.astype(np.uint8)
    summary_writer.image(tag='2dimage_test', image=img, step=1)
    summary_value = self.parse_and_return_summary_value(path=log_dir)

    self.assertEqual(summary_value.tag, '2dimage_test')
    actual_img = tf.image.decode_image(summary_value.tensor.string_val[2])
    # assert the image was increased in dimension
    self.assertEqual(actual_img.shape, (30, 30, 3))

  def test_summarywriter_single_channel_image_scaled(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    img = np.random.uniform(low=0.0, high=255.0, size=(30, 30, 1))
    img = img.astype(np.uint8)
    summary_writer.image(tag='2dimage_1channel_test', image=img, step=1)
    summary_value = self.parse_and_return_summary_value(path=log_dir)

    self.assertEqual(summary_value.tag, '2dimage_1channel_test')
    actual_img = tf.image.decode_image(summary_value.tensor.string_val[2])
    # assert the image was increased in dimension
    self.assertEqual(actual_img.shape, (30, 30, 3))

  def test_summarywriter_multiple_images(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    expected_img = np.random.uniform(low=0.0, high=255.0, size=(2, 30, 30, 3))
    expected_img = expected_img.astype(np.uint8)
    summary_writer.image(tag='multiple_images_test', image=expected_img, step=1)
    summary_value = self.parse_and_return_summary_value(path=log_dir)

    self.assertEqual(summary_value.tag, 'multiple_images_test')
    actual_imgs = [
      tf.image.decode_image(s) for s in summary_value.tensor.string_val[2:]
    ]
    self.assertTrue(np.allclose(np.stack(actual_imgs, axis=0), expected_img))

  def test_summarywriter_multiple_2dimages_scaled(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    img = np.random.uniform(low=0.0, high=255.0, size=(2, 30, 30))
    img = img.astype(np.uint8)
    summary_writer.image(tag='multiple_2dimages_test', image=img, step=1)
    summary_value = self.parse_and_return_summary_value(path=log_dir)

    self.assertEqual(summary_value.tag, 'multiple_2dimages_test')
    actual_imgs = [
      tf.image.decode_image(s) for s in summary_value.tensor.string_val[2:]
    ]
    # assert the images were increased in dimension
    self.assertEqual(np.stack(actual_imgs, axis=0).shape, (2, 30, 30, 3))

  def test_summarywriter_audio(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    expected_audio_samples = np.random.uniform(
      low=-1.0, high=1.0, size=(2, 48000, 2)
    )
    summary_writer.audio(
      tag='audio_test', audiodata=expected_audio_samples, step=1
    )

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'audio_test')

    # Assert two audio files are parsed.
    self.assertLen(summary_value.tensor.string_val, 2)

    # Assert values.
    actual_audio_1 = tf.audio.decode_wav(
      summary_value.tensor.string_val[0]
    ).audio
    self.assertTrue(
      np.allclose(expected_audio_samples[0], actual_audio_1, atol=1e-04)
    )

    actual_audio_2 = tf.audio.decode_wav(
      summary_value.tensor.string_val[1]
    ).audio
    self.assertTrue(
      np.allclose(expected_audio_samples[1], actual_audio_2, atol=1e-04)
    )

  def test_summarywriter_audio_sampled_output(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    expected_audio_samples = np.random.uniform(
      low=-1.0, high=1.0, size=(2, 48000, 2)
    )
    summary_writer.audio(
      tag='audio_test',
      audiodata=expected_audio_samples,
      step=1,
      max_outputs=1,
    )

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'audio_test')

    # Assert only the first audio clip is available.
    self.assertLen(summary_value.tensor.string_val, 1)

    # Assert values.
    actual_audio = tf.audio.decode_wav(summary_value.tensor.string_val[0]).audio

    self.assertTrue(
      np.allclose(expected_audio_samples[0], actual_audio, atol=1e-04)
    )

  def test_summarywriter_clipped_audio(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    expected_audio_samples = np.random.uniform(
      low=-2.0, high=2.0, size=(2, 48000, 2)
    )
    summary_writer.audio(
      tag='audio_test',
      audiodata=expected_audio_samples,
      step=1,
      max_outputs=1,
    )

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'audio_test')

    # Assert one audio files is parsed.
    self.assertLen(summary_value.tensor.string_val, 1)

    # actual_audio is clipped.
    actual_audio = tf.audio.decode_wav(summary_value.tensor.string_val[0]).audio
    self.assertFalse(
      np.allclose(expected_audio_samples[0], actual_audio, atol=1e-04)
    )

    clipped_audio = np.clip(np.array(expected_audio_samples[0]), -1, 1)
    self.assertTrue(np.allclose(clipped_audio, actual_audio, atol=1e-04))

  def test_summarywriter_histogram_defaultbins(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    histogram = np.arange(1000)
    # Histogram will be created for 30 (default) bins.
    summary_writer.histogram(tag='histogram_test', values=histogram, step=1)

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'histogram_test')
    actual_histogram = tensor_util.make_ndarray(summary_value.tensor)
    self.assertTrue(actual_histogram.shape, (30, 3))
    self.assertTrue(
      np.allclose(actual_histogram[0], (0.0, 33.3, 34.0), atol=1e-01)
    )

  def test_summarywriter_histogram_2bins(self):
    log_dir = tempfile.mkdtemp()
    summary_writer = SummaryWriter(log_dir=log_dir)
    histogram = np.arange(1000)
    summary_writer.histogram(
      tag='histogram_test', values=histogram, step=1, bins=2
    )

    summary_value = self.parse_and_return_summary_value(path=log_dir)
    self.assertEqual(summary_value.tag, 'histogram_test')
    actual_histogram = tensor_util.make_ndarray(summary_value.tensor)
    self.assertTrue(actual_histogram.shape, (2, 3))
    self.assertTrue(
      np.allclose(actual_histogram[0], (0.0, 499.5, 500.0), atol=1e-01)
    )
    self.assertTrue(
      np.allclose(actual_histogram[1], (499.5, 999.0, 500.0), atol=1e-01)
    )

  def test_flatten_dict(self):
    # Valid types according to https://github.com/tensorflow/tensorboard/blob/1204566da5437af55109f7a4af18f9f8b7c4f864/tensorboard/plugins/hparams/summary_v2.py
    input_hparams = {
      # Example Invalid Types
      'None': None,
      'List': [1, 2, 3],
      'Tuple': (1, 2, 3),
      'Complex': complex('1+1j'),
      'np.complex_': np.complex_('1+1j'),
      # Valid Python Types
      'Bool': True,
      'Int': 1,
      'Float': 1.0,
      'Str': 'test',
      # Valid Numpy Types
      'np.bool_': np.bool_(1),
      'np.integer': np.int_(1),
      'np.floating': np.float_(1.0),
      'np.character': np.str_('test'),
      # Nested dict to flatten
      'Nested_Dict': {
        'None': None,
        'List': [1, 2, 3],
        'Tuple': (1, 2, 3),
        'Complex': complex('1+1j'),
        'np.complex_': np.complex_('1+1j'),
        'Bool': True,
        'Int': 1,
        'Float': 1.0,
        'Str': 'test',
        'np.bool_': np.bool_(1),
        'np.integer': np.int_(1),
        'np.floating': np.float_(1.0),
        'np.character': np.str_('test'),
      },
    }

    result_hparams = _flatten_dict(input_hparams)

    expected_hparams = {
      'None': 'None',
      'List': '[1, 2, 3]',
      'Tuple': '(1, 2, 3)',
      'Complex': '(1+1j)',
      'np.complex_': '(1+1j)',
      # Valid Python Types
      'Bool': True,
      'Int': 1,
      'Float': 1.0,
      'Str': 'test',
      # Valid Numpy Types
      'np.bool_': np.bool_(1),
      'np.integer': np.int_(1),
      'np.floating': np.float_(1.0),
      'np.character': np.str_('test'),
      # Nested Dict
      'Nested_Dict.None': 'None',
      'Nested_Dict.List': '[1, 2, 3]',
      'Nested_Dict.Tuple': '(1, 2, 3)',
      'Nested_Dict.Complex': '(1+1j)',
      'Nested_Dict.np.complex_': '(1+1j)',
      'Nested_Dict.Bool': True,
      'Nested_Dict.Int': 1,
      'Nested_Dict.Float': 1.0,
      'Nested_Dict.Str': 'test',
      'Nested_Dict.np.bool_': np.bool_(1),
      'Nested_Dict.np.integer': np.int_(1),
      'Nested_Dict.np.floating': np.float_(1.0),
      'Nested_Dict.np.character': np.str_('test'),
    }

    self.assertDictEqual(result_hparams, expected_hparams)

  def test_auto_flush(self):
    tmp_dir = pathlib.Path(self.create_tempdir().full_path)
    summary_writer = SummaryWriter(tmp_dir, auto_flush=True)
    summary_writer.scalar('metric', 123, 1)
    filesize_before_flush = _disk_usage(tmp_dir)
    summary_writer.flush()
    filesize_after_flush = _disk_usage(tmp_dir)
    self.assertEqual(filesize_before_flush, filesize_after_flush)

  def test_no_auto_flush(self):
    tmp_dir = pathlib.Path(self.create_tempdir().full_path)
    summary_writer = SummaryWriter(tmp_dir, auto_flush=False)
    summary_writer.scalar('metric', 123, 1)
    filesize_before_flush = _disk_usage(tmp_dir)
    summary_writer.flush()
    filesize_after_flush = _disk_usage(tmp_dir)
    self.assertLess(filesize_before_flush, filesize_after_flush)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import jax
import jax.numpy as jnp

from flax.experimental import nnx


class Block(nnx.Module):
  def __init__(self, dim: int, *, rngs: nnx.Rngs):
    self.linear = nnx.Linear(dim, dim, rngs=rngs)
    self.dropout = nnx.Dropout(0.5)

  def __call__(self, x: jax.Array, *, rngs: nnx.Rngs) -> jax.Array:
    x = self.linear(x)
    x = self.dropout(x, rngs=rngs)
    x = jax.nn.gelu(x)
    return x


class ScanMLP(nnx.Module):
  """
  An MLP that uses `vmap` during `__init__` to create a Block instance
  with an additional `layer` axis, and `scan` during `__call__` to apply
  the sequence of layers iteratively over the input / output `x`.
  """

  def __init__(self, dim: int, *, n_layers: int, rngs: nnx.Rngs):
    self.n_layers = n_layers
    # fork Rngs, split keys into `n_layers`
    keys = rngs.fork(n_layers)

    def create_block(keys):
      # create Block instance and return its split
      return Block(dim, rngs=nnx.Rngs(keys)).split()

    # call vmap over create_block, passing the split `params` key
    # and immediately merge to get a Block instance
    self.layers = nnx.merge(jax.vmap(create_block)(keys))

  def __call__(self, x: jax.Array, *, rngs: nnx.Rngs) -> jax.Array:
    # fork Rngs, split keys into `n_layers`
    keys = rngs.fork(self.n_layers)
    # split Module to get params
    params, static = self.layers.split(nnx.Param)

    def scan_fn(
      x: jax.Array, inputs: Tuple[nnx.State, dict[str, nnx.RngStream]]
    ) -> Tuple[jax.Array, nnx.State]:
      params, keys = inputs
      # merge back Module and Rngs
      module = static.merge(params)
      # forward pass
      x = module(x, rngs=nnx.Rngs(keys))
      # split state and return
      params, _ = module.split(nnx.Param)
      return x, params

    # call scan passing x as the carry, and params + keys as the input
    x, params = jax.lax.scan(scan_fn, x, (params, keys))
    # update layers state and return
    self.layers.update(params)
    return x


model = ScanMLP(10, n_layers=5, rngs=nnx.Rngs(0))

x = jnp.ones((3, 10))
with nnx.flags(deterministic=False):
  y = model(x, rngs=nnx.Rngs(dropout=1))

print(jax.tree_map(jnp.shape, model.get_state()))
print(y.shape)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from flax.experimental import nnx


class TestContainers:
  def test_node_idenpotence(self):
    x = nnx.Variable(1)
    x = nnx.Variable(x)

    assert isinstance(x, nnx.Variable)

  def test_variable_idenpotence(self):
    x = nnx.Variable(1)
    x = nnx.Variable(x)

    assert isinstance(x, nnx.Variable)
    assert x.value == 1

  def test_variable_cannot_change_collection(self):
    x = nnx.Param(1)

    with pytest.raises(ValueError, match='is not compatible with return type'):
      x = nnx.BatchStat(x)

  def test_container_cannot_change_type(self):
    x = nnx.Variable(1)

    with pytest.raises(ValueError, match='is not compatible with return type'):
      x = nnx.Param(x)

    x = nnx.Param(2)

    with pytest.raises(ValueError, match='is not compatible with return type'):
      x = nnx.Variable(x)

  def test_unbox(self):
    x: nnx.Param[int] = nnx.Param(
      1,
      get_value_hooks=[lambda c, x: x + 1, lambda c, x: x * 2],  # type: ignore
    )

    assert x.get_value() == 4

  def test_box(self):
    x: nnx.Param[int] = nnx.Param(
      1,  # type: ignore
      set_value_hooks=[lambda c, x: x + 1, lambda c, x: x * 2],  # type: ignore
    )
    x.set_value(5)

    assert x.value == 12

  def test_module_unbox(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.x = nnx.Param(
          1, get_value_hooks=[lambda c, x: x + 1, lambda c, x: x * 2]
        )

    module = Foo()

    assert module.x == 4
    assert vars(module)['x'].value == 1

  def test_module_box(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.x = nnx.Param(
          1, set_value_hooks=[lambda c, x: x + 1, lambda c, x: x * 2]
        )

    module = Foo()
    module.x = 5

    assert module.x == 12
    assert vars(module)['x'].value == 12



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.examples.seq2seq.train."""

import functools

from absl.testing import absltest
from flax.training import train_state
import jax
from jax import random
import numpy as np
import optax

import input_pipeline
import train
import models

jax.config.parse_flags_with_absl()


def create_ctable(chars='0123456789+= '):
  return input_pipeline.CharacterTable(chars)


def create_train_state(ctable):
  model = models.Seq2seq(
      teacher_force=False,
      hidden_size=train.FLAGS.hidden_size,
      vocab_size=ctable.vocab_size,
  )
  params = train.get_initial_params(model, jax.random.key(0), ctable)
  tx = optax.adam(train.FLAGS.learning_rate)
  state = train_state.TrainState.create(
      apply_fn=model.apply, params=params, tx=tx
  )
  return state


class TrainTest(absltest.TestCase):

  def test_character_table(self):
    ctable = create_ctable()
    text = '410+19'
    enc_text = ctable.encode(text)
    dec_text = ctable.decode(enc_text)
    # The text is possibly padded with whitespace, but the trimmed output should
    # be equal to the input.
    self.assertEqual(text, dec_text.strip())

  def test_mask_sequences(self):
    np.testing.assert_equal(
        input_pipeline.mask_sequences(
            np.arange(1, 13).reshape((4, 3)), np.array([3, 2, 1, 0])
        ),
        np.array([[1, 2, 3], [4, 5, 0], [7, 0, 0], [0, 0, 0]]),
    )

  def test_get_sequence_lengths(self):
    oh_sequence_batch = jax.vmap(
        functools.partial(jax.nn.one_hot, num_classes=4)
    )(np.array([[0, 1, 0], [1, 0, 2], [1, 2, 0], [1, 2, 3]]))
    np.testing.assert_equal(
        input_pipeline.get_sequence_lengths(oh_sequence_batch, eos_id=0),
        np.array([1, 2, 3, 3], np.int32),
    )
    np.testing.assert_equal(
        input_pipeline.get_sequence_lengths(oh_sequence_batch, eos_id=1),
        np.array([2, 1, 1, 1], np.int32),
    )
    np.testing.assert_equal(
        input_pipeline.get_sequence_lengths(oh_sequence_batch, eos_id=2),
        np.array([3, 3, 2, 2], np.int32),
    )

  def test_train_one_step(self):
    ctable = create_ctable()
    batch = ctable.get_batch(128)

    state = create_train_state(ctable)
    key = random.key(0)
    _, train_metrics = train.train_step(state, batch, key, ctable.eos_id)

    self.assertLessEqual(train_metrics['loss'], 5)
    self.assertGreaterEqual(train_metrics['accuracy'], 0)

  def test_decode_batch(self):
    ctable = create_ctable()
    batch = ctable.get_batch(5)
    key = random.key(0)
    state = create_train_state(ctable)
    train.decode_batch(state, batch, key, ctable)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

import jax
import numpy as np
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax import errors
from flax.core import FrozenDict, apply, copy, init, lift, nn


class LiftTest(absltest.TestCase):
  def test_aliasing(self):
    def f(scope):
      a = scope.push('a')

      def g(scopes, _):
        scope, a = scopes
        self.assertEqual(a.parent, scope)

      lift.vmap(g, variable_axes={}, split_rngs={})((scope, a), jnp.ones((1,)))

    init(f)(random.key(0))

  def test_undefined_param(self):
    def f(scope):
      dense = lift.vmap(
        nn.dense,
        in_axes=(0, None),
        out_axes=0,
        variable_axes={'params': 0},
        split_rngs={'params': True},
      )
      dense(scope.push('dense'), np.ones((3, 2)), 2)

    msg = r'Could not find parameter named "kernel" in scope "/vmap\(dense\)".'
    with self.assertRaisesRegex(errors.ScopeParamNotFoundError, msg):
      apply(f)({'params': {'dense': {'abc': np.ones((3, 3))}}})

  def test_jit_cache(self):
    compiles = 0

    @lift.jit
    def f(scope, _module_hash, x):
      nonlocal compiles
      compiles += 1
      if scope.is_mutable_collection(
        'intermediates'
      ) and not scope.is_mutable_collection('params'):
        scope.put_variable('intermediates', 'x', x + 1)
      return nn.dense(scope, x, 1)

    x = np.ones((3, 2))
    module_hash = 1
    _, params = init(f)(random.key(0), module_hash, x)
    init(f)(random.key(0), module_hash, x)
    self.assertEqual(compiles, 1)
    apply(f)(params, module_hash, x)
    self.assertEqual(compiles, 2)  # apply should cause a compile
    apply(f)(params, module_hash, x)
    self.assertEqual(compiles, 2)  # applying again should not
    # edge case where only the implicit return of the jitted functions changes.
    # this should not use the previously cached apply.
    _, state = apply(f, mutable='intermediates')(params, module_hash, x)
    self.assertEqual(compiles, 3)  # applying again should not
    self.assertEqual(state['intermediates']['x'].sum(), 3 * 2 * 2)

  def test_vjp(self):
    def g(scope, x, y):
      p = scope.param('test', nn.initializers.constant(0.5), ())
      scope.variable('state', 'counter', lambda: 0)
      return p * x * y

    def f(scope, x, y):
      z, bwd = lift.vjp(g, scope, x, y)
      return bwd(jnp.ones(y.shape))

    x = jnp.array([1.0, 2.0, 3.0])
    y = jnp.array([4.0, 5.0, 6.0])
    _, params = init(f)(random.key(0), x, y)
    params_grad, x_grad, y_grad = apply(f)(params, x, y)
    self.assertEqual(
      params_grad,
      {
        'params': FrozenDict({'test': 32.0}),
      },
    )
    np.testing.assert_allclose(x_grad, [2.0, 2.5, 3.0])
    np.testing.assert_allclose(y_grad, [0.5, 1.0, 1.5])

  def test_jvp(self):
    def g(scope, x):
      p = scope.param('test', nn.initializers.zeros_init(), ())
      scope.variable('state', 'counter', lambda: 0)
      return p * x

    def f(scope, x):
      vars_t = jax.tree_util.tree_map(
        jnp.ones_like, scope.variables().get('params', {})
      )
      _, out_t = lift.jvp(
        g, scope, (x,), (jnp.zeros_like(x),), {'params': vars_t}
      )
      return out_t

    x = jnp.ones((3,))
    _, params = init(f)(random.key(0), x)
    y_t = apply(f)(params, x)
    np.testing.assert_allclose(y_t, jnp.ones_like(x))

  def test_while_loop(self):
    def f(scope, x):
      key_zero = random.key(0)
      key_zero = jnp.broadcast_to(key_zero, (2, *key_zero.shape))
      scope.param('inc', lambda _: 1)
      scope.put_variable('state', 'acc', 0)
      scope.put_variable('state', 'rng_params', key_zero)
      scope.put_variable('state', 'rng_loop', key_zero)

      def cond_fn(scope, c):
        acc = scope.get_variable('state', 'acc')
        return acc < x

      def body_fn(scope, c):
        i = scope.get_variable('state', 'acc')
        p_rng = scope.make_rng('params')
        l_rng = scope.make_rng('loop')
        scope.put_variable(
          'state',
          'rng_params',
          scope.get_variable('state', 'rng_params').at[i].set(p_rng),
        )
        scope.put_variable(
          'state',
          'rng_loop',
          scope.get_variable('state', 'rng_loop').at[i].set(l_rng),
        )
        inc = scope.get_variable('params', 'inc')
        scope.put_variable('state', 'acc', i + inc)
        return c + 2

      return lift.while_loop(
        cond_fn,
        body_fn,
        scope,
        0,
        carry_variables='state',
        split_rngs={'params': False, 'loop': True},
      )

    x = 2
    c, vars = apply(f, mutable=True)(
      {}, x, rngs={'params': random.key(1), 'loop': random.key(2)}
    )
    self.assertEqual(vars['state']['acc'], x)
    self.assertEqual(c, 2 * x)
    np.testing.assert_array_equal(
      vars['state']['rng_params'][0], vars['state']['rng_params'][1]
    )
    np.testing.assert_array_compare(
      operator.__ne__,
      vars['state']['rng_loop'][0],
      vars['state']['rng_loop'][1],
    )

  def test_cond(self):
    def f(scope, x, pred):
      scope.variable('state', 'true_count', lambda: 0)
      scope.variable('state', 'false_count', lambda: 0)

      def true_fn(scope, x):
        scope.variable('state', 'true_count').value += 1
        return scope.child(nn.dense)(x, 2)

      def false_fn(scope, x):
        scope.variable('state', 'false_count').value += 1
        return -scope.child(nn.dense)(x, 2)

      return lift.cond(pred, true_fn, false_fn, scope, x)

    x = jnp.ones((1, 3))
    y1, vars = init(f)(random.key(0), x, True)
    self.assertEqual(vars['state'], {'true_count': 1, 'false_count': 0})
    y2, vars = apply(f, mutable='state')(vars, x, False)
    self.assertEqual(vars['state'], {'true_count': 1, 'false_count': 1})
    np.testing.assert_allclose(y1, -y2)

  def test_switch(self):
    def f(scope, x, index):
      scope.variable('state', 'a_count', lambda: 0)
      scope.variable('state', 'b_count', lambda: 0)
      scope.variable('state', 'c_count', lambda: 0)

      def a_fn(scope, x):
        scope.variable('state', 'a_count').value += 1
        return scope.child(nn.dense)(x, 2)

      def b_fn(scope, x):
        scope.variable('state', 'b_count').value += 1
        return -scope.child(nn.dense)(x, 2)

      def c_fn(scope, x):
        scope.variable('state', 'c_count').value += 1
        return scope.child(nn.dense)(x, 2)

      return lift.switch(index, [a_fn, b_fn, c_fn], scope, x)

    x = jnp.ones((1, 3))
    y1, vars = init(f)(random.key(0), x, 0)
    self.assertEqual(vars['state'], {'a_count': 1, 'b_count': 0, 'c_count': 0})
    y2, updates = apply(f, mutable='state')(vars, x, 1)
    vars = copy(vars, updates)
    self.assertEqual(vars['state'], {'a_count': 1, 'b_count': 1, 'c_count': 0})
    np.testing.assert_allclose(y1, -y2)
    y3, updates = apply(f, mutable='state')(vars, x, 2)
    vars = copy(vars, updates)
    self.assertEqual(vars['state'], {'a_count': 1, 'b_count': 1, 'c_count': 1})
    np.testing.assert_allclose(y1, y3)

  def test_subscope_var_aliasing(self):
    def test(scope, x):
      subscope = scope.push(name='a')
      subscope.put_variable('state', 'x', 0.0)
      _ = lift.while_loop(
        lambda scope, x: False,
        lambda scope, x: x,
        scope,
        jnp.array(0, jnp.int32),
        carry_variables=['state'],
      )
      subscope.put_variable('state', 'x', 1.0)
      val0 = scope.variables()['state']['a']['x']
      val1 = subscope.variables()['state']['x']
      self.assertEqual(val0, val1)
      return x

    init(test)(random.key(0), 1.0)


if __name__ == '__main__':
  absltest.main()
